import os
from os.path import join as opj
from datetime import datetime
import datetime
import urllib
from typing import Union

import requests
import random
import json
import pandas as pd
import time
import statistics
from progressbar import ProgressBar
import progressbar
import numpy as np
import string
import io
import threading


def generate_unique_name(base_name:str, length=6):
    # TODO this function has 4 duplicates
    random_suffix = ''.join(random.choices(string.ascii_lowercase, k=length))
    unique_name = base_name + '_' + random_suffix
    return unique_name


# from US_states_abbreviation import getabreviation
def get_iem_bystation(dates:Union[tuple, list], station:str, path:os.PathLike, mettag:str):
    '''
      Dates is a tupple/list of strings with date ranges
      
      an example date string should look like this: dates = ["01-01-2012","12-31-2012"]
      
      if station is given data will be downloaded directly from the station the default is false.
      
      mettag: your prefered suffix to save on filee
      '''
    # access the elements in the metdate class above
    wdates = metdate(dates)
    stationx = station[:2]
    state_clim = stationx + "CLIMATE"
    str0 = "http://mesonet.agron.iastate.edu/cgi-bin/request/coop.py?network="
    str1 = str0 + state_clim + "&stations=" + station
    str2 = str1 + "&year1=" + wdates.year_start + "&month1=" + wdates.startmonth + "&day1=" + wdates.startday + "&year2=" + wdates.year_end + "&month2=" + wdates.endmonth + "&day2=" + wdates.endday
    str3 = (str2 + "&vars%5B%5D=apsim&what=view&delim=comma&gis=no")
    url = str3
    rep = requests.get(url)
    if rep.ok:
        metname = station + mettag + ".met"
        os.chdir(path)
        if not os.path.exists('weatherdata'):
            os.mkdir('weatherdata')
        pt = os.path.join('weatherdata', metname)

        with open(pt, 'wb') as metfile1:
            # dont forget to include the file extension name
            metfile1.write(rep.content)
            rep.close()
            metfile1.close()
            print(rep.content)
    else:
        print("Failed to download the data web request returned code: ", rep)


class metdate:
    def __init__(self, dates):
        self.startdate = dates[0]
        self.lastdate = dates[1]
        self.startmonth = dates[0][:2]
        self.endmonth = dates[1][:2]
        self.year_start = dates[0].split("-")[2]
        self.year_end = dates[1].split("-")[2]
        self.startday = datetime.datetime.strptime(dates[0], '%m-%d-%Y').strftime('%j')
        self.endday = datetime.datetime.strptime(dates[1], '%m-%d-%Y').strftime('%j')


dates = ['01-01-2000', '12-31-2020']

states = {
    'AK': 'Alaska',
    'AL': 'Alabama',
    'AR': 'Arkansas',
    'AZ': 'Arizona',
    'CA': 'California',
    'CO': 'Colorado',
    'CT': 'Connecticut',
    'DC': 'District of Columbia',
    'DE': 'Delaware',
    'FL': 'Florida',
    'GA': 'Georgia',
    'HI': 'Hawaii',
    'IA': 'Iowa',
    'ID': 'Idaho',
    'IL': 'Illinois',
    'IN': 'Indiana',
    'KS': 'Kansas',
    'KY': 'Kentucky',
    'LA': 'Louisiana',
    'MA': 'Massachusetts',
    'MD': 'Maryland',
    'ME': 'Maine',
    'MI': 'Michigan',
    'MN': 'Minnesota',
    'MO': 'Missouri',
    'MS': 'Mississippi',
    'MT': 'Montana',
    'NC': 'North Carolina',
    'ND': 'North Dakota',
    'NE': 'Nebraska',
    'NH': 'New Hampshire',
    'NJ': 'New Jersey',
    'NM': 'New Mexico',
    'NV': 'Nevada',
    'NY': 'New York',
    'OH': 'Ohio',
    'OK': 'Oklahoma',
    'OR': 'Oregon',
    'PA': 'Pennsylvania',
    'RI': 'Rhode Island',
    'SC': 'South Carolina',
    'SD': 'South Dakota',
    'TN': 'Tennessee',
    'TX': 'Texas',
    'UT': 'Utah',
    'VA': 'Virginia',
    'VT': 'Vermont',
    'WA': 'Washington',
    'WI': 'Wisconsin',
    'WV': 'West Virginia',
    'WY': 'Wyoming'
}

# flip the keys
new_dict = {}
for k, v in states.items():
    new_dict[v] = k


def getabreviation(x):
    ab = new_dict[x]
    return (ab)


# function to define the date ranges
def daterange(start, end):
    '''
  start: the starting year to download the weather data
  -----------------
  end: the year under which download should stop
  '''
    startdates = '01-01'
    enddates = '12-31'
    end = str(end) + "-" + enddates
    start = str(start) + "-" + startdates
    drange = pd.date_range(start, end)
    return (drange)


# check if a year is aleap year
def isleapyear(year):
    if (year % 400 == 0) and (year % 100 == 0) or (year % 4 == 0) and (year % 100 != 0):
        return (True)
    else:
        return (False)


# download radiation data for replacement
def get_nasarad(lonlat:Union[tuple, list], start:int, end:int):
    lon = lonlat[0]
    lat = lonlat[1]
    pars = "ALLSKY_SFC_SW_DWN"
    rm = f'https://power.larc.nasa.gov/api/temporal/daily/point?start={start}0101&end={end}1231&latitude={lat}&longitude={lon}&community=ag&parameters={pars}&format=json&user=richard&header=true&time-standard=lst'
    data = requests.get(rm)
    dt = json.loads(data.content)
    df = pd.DataFrame(dt["properties"]['parameter'])
    if len(df) == len(daterange(start, end)):
        return df

    # fucntion to download data from daymet


def daymet_bylocation(lonlat:Union[tuple, list], start:int, end:int, cleanup:bool=True, filename:str=None):
    '''collect weather from daymet solar radiation is replaced with that of nasapower
   ------------
   parameters
   ---------------
   start: Starting year
   
   end: Ending year
   
   lonlat: A tuple of xy cordnates
   
   Cleanup:  A bolean True or False default is true: deletes the excel file generated during the file write up
   
   ------------
   returns complete path to the new met file but also write the met file to the disk in the working dir_path
   '''
    # import pdb
    # pdb.set_trace()

    datecheck = daterange(start, end)
    if start < 1980 or end > 2021:
        print(
            "requested year preceeds valid data range! \n end years should not exceed 2021 and start year should not be less than 1980")
    else:
        base_url = 'https://daymet.ornl.gov/single-pixel/api/data?'
        latstr = 'lat=' + str(lonlat[1])
        lonstr = '&lon=' + str(lonlat[0])
        varss = ['dayl', 'prcp', 'srad', 'tmax', 'tmin', 'vp', 'swe']
        setyears = [str(year) for year in range(start, end + 1)]
        # join the years as a string
        years_in_range = ",".join(setyears)
        years_str = "&years=" + years_in_range
        varfield = ",".join(varss)
        var_str = "&measuredParams=" + varfield
        # join the string url together
        url = base_url + latstr + lonstr + var_str + years_str
        conn = requests.get(url, timeout=60)

        if not conn.ok:
            print("failed to connect to server")
        elif conn.ok:
            # print("connection established to download the following data", url)
            # outFname = conn.headers["Content-Disposition"].split("=")[-1]
            outFname = conn.headers["Content-Disposition"].split("=")[-1]
            text_str = conn.content
            outF = open(outFname, 'wb')
            outF.write(text_str)
            outF.close()
            conn.close()
            #       read the downloaded data to a data frame
            dmett = pd.read_csv(outFname, delimiter=',', skiprows=7)
            vp = dmett['vp (Pa)'] * 0.01
            # calcuate radn
            radn = dmett['dayl (s)'] * dmett['srad (W/m^2)'] * 1e-06
            # re-arrange data frame
            year = np.array(dmett['year'])
            day = np.array(dmett['yday'])
            radn = np.array(radn)
            maxt = np.array(dmett['tmax (deg c)'])
            mint = np.array(dmett['tmin (deg c)'])
            rain = np.array(dmett['prcp (mm/day)'])
            vp = np.array(vp)
            swe = np.array(dmett['swe (kg/m^2)'])
            df = pd.DataFrame(
                {'year': year, 'day': day, 'radn': radn, 'maxt': maxt, 'mint': mint, 'rain': rain, 'vp': vp,
                 'swe': swe})
            # bind the frame
            # calculate mean annual applitude in mean monthly temperature (TAV)
            ab = [a for a in setyears]
            # split the data frame
            ab = [x for _, x in df.groupby(df['year'])]
            df_bag = []
            # constants to evaluate the leap years
            leapfactor = 4
            for i in ab:
                if (all(i.year % 400 == 0)) and (all(i.year % 100 == 0)) or (all(i.year % 4 == 0)) and (
                all(i.year % 100 != 0)):
                    x = i[['year', 'radn', 'maxt', 'mint', 'rain', 'vp', 'swe', ]].mean()
                    year = round(x.iloc[0], 0)
                    day = round(366, 0)
                    new_row = pd.DataFrame(
                        {'year': [year], 'day': [day], 'radn': [0], 'maxt': [0], 'mint': [0], 'rain': [0], 'vp': [0],
                         'swe': [0]})
                    df_bag.append(pd.concat([i, new_row], ignore_index=True))

                    continue
                else:
                    df_bag.append(i)
                    frames = df_bag
            newmet = pd.concat(frames)
            newmet.index = range(0, len(newmet))
            # repalce radn data
            rad = get_nasarad(lonlat, start, end)
            newmet["radn"] = rad.ALLSKY_SFC_SW_DWN.values
            if len(newmet) != len(datecheck):
                print('date discontinuities still exisists')
            else:
                # print("met data is in the range of specified dates no discontinuities")
                rg = len(newmet.day.values) + 1
                # newmet  = pd.concat(newmet)
                mean_maxt = newmet['maxt'].mean(skipna=True, numeric_only=None)
                mean_mint = newmet['mint'].mean(skipna=True, numeric_only=None)
                AMP = round(mean_maxt - mean_mint, 2)
                tav = round(statistics.mean((mean_maxt, mean_mint)), 2)
                tile = conn.headers["Content-Disposition"].split("=")[1].split("_")[0]
                fn = conn.headers["Content-Disposition"].split("=")[1].replace("csv", 'met')
                if not filename:
                    shortenfn = generate_unique_name("Daymet") + '.met'
                else:
                    shortenfn = filename
                if not os.path.exists('weatherdata'):
                    os.makedirs('weatherdata')
                fn = shortenfn
                fname = os.path.join('weatherdata', fn)
                headers = ['year', 'day', 'radn', 'maxt', 'mint', 'rain', 'vp', 'swe']
                header_string = " ".join(headers) + "\n"
                # close and append new lines
                with open(fname, "a") as f2app:
                    f2app.writelines([f'!site: {tile}\n', f'latitude = {lonlat[1]} \n', f'longitude = {lonlat[0]}\n',
                                      f'tav ={tav}\n', f'amp ={AMP}\n'])
                    f2app.writelines([header_string])
                    f2app.writelines(['() () (MJ/m2/day) (oC) (oC) (mm) (hPa) (kg/m2)\n'])
                    # append the weather data
                    data_rows = []
                    for index, row in newmet.iterrows():
                        current_row = []
                        for header in headers:
                            current_row.append(str(row[header]))
                        current_str = " ".join(current_row) + '\n'
                        data_rows.append(current_str)

                    f2app.writelines(data_rows)

                if cleanup:
                    if os.path.isfile(os.path.join(os.getcwd(), outFname)):
                        os.remove(os.path.join(os.getcwd(), outFname))
                return fname  # fname


def daymet_bylocation_nocsv(lonlat:Union[tuple, list], start:int, end:int, cleanup:bool=True, filename:str=None):
    '''collect weather from daymet solar radiation is replaced with that of nasapower
    ------------
    parameters
    ---------------
    start: Starting year

    end: Ending year

    lonlat: A tuple of xy cordnates

    Cleanup:  A bolean True or False default is true: deletes the excel file generated during the file write up

    ------------
    returns a complete path to the new met file but also write the met file to the disk in the working dir_path
    '''
    # import pdb
    # pdb.set_trace()

    datecheck = daterange(start, end)
    if start < 1980 or end > 2021:
        print(
            "requested year preceeds valid data range! \n end years should not exceed 2021 and start year should not be less than 1980")
    else:
        base_url = 'https://daymet.ornl.gov/single-pixel/api/data?'
        latstr = 'lat=' + str(lonlat[1])
        lonstr = '&lon=' + str(lonlat[0])
        varss = ['dayl', 'prcp', 'srad', 'tmax', 'tmin', 'vp', 'swe']
        setyears = [str(year) for year in range(start, end + 1)]
        # join the years as a string
        years_in_range = ",".join(setyears)
        years_str = "&years=" + years_in_range
        varfield = ",".join(varss)
        var_str = "&measuredParams=" + varfield
        # join the string url together
        url = base_url + latstr + lonstr + var_str + years_str
        conn = requests.get(url, timeout=60)

        if not conn.ok:
            print("failed to connect to server")
        elif conn.ok:
            # print("connection established to download the following data", url)
            # outFname = conn.headers["Content-Disposition"].split("=")[-1]
            outFname = "w" + filename + conn.headers["Content-Disposition"].split("=")[-1]
            text_str = conn.content
            conn.close()
            #  read the downloaded data to a data frame
            # Create an in-memory binary stream
            text_stream = io.BytesIO(text_str)
            # Read the data into a DataFrame
            dmett = pd.read_csv(text_stream, delimiter=',', skiprows=7)
            # dmett = pd.read_csv(outFname, delimiter=',', skiprows=7)
            vp = dmett['vp (Pa)'] * 0.01
            # calcuate radn
            radn = dmett['dayl (s)'] * dmett['srad (W/m^2)'] * 1e-06
            # re-arrange data frame
            year = np.array(dmett['year'])
            day = np.array(dmett['yday'])
            radn = np.array(radn)
            maxt = np.array(dmett['tmax (deg c)'])
            mint = np.array(dmett['tmin (deg c)'])
            rain = np.array(dmett['prcp (mm/day)'])
            vp = np.array(vp)
            swe = np.array(dmett['swe (kg/m^2)'])
            df = pd.DataFrame(
                {'year': year, 'day': day, 'radn': radn, 'maxt': maxt, 'mint': mint, 'rain': rain, 'vp': vp,
                 'swe': swe})
            # bind the frame
            # calculate mean annual applitude in mean monthly temperature (TAV)
            ab = [a for a in setyears]
            # split the data frame
            ab = [x for _, x in df.groupby(df['year'])]
            df_bag = []
            # constants to evaluate the leap years
            leapfactor = 4
            for i in ab:
                if (all(i.year % 400 == 0)) and (all(i.year % 100 == 0)) or (all(i.year % 4 == 0)) and (
                        all(i.year % 100 != 0)):
                    x = i[['year', 'radn', 'maxt', 'mint', 'rain', 'vp', 'swe', ]].mean()
                    year = round(x[0], 0)
                    day = round(366, 0)
                    new_row = pd.DataFrame(
                        {'year': [year], 'day': [day], 'radn': [0], 'maxt': [0], 'mint': [0], 'rain': [0], 'vp': [0],
                         'swe': [0]})
                    df_bag.append(pd.concat([i, new_row], ignore_index=True))

                    continue
                else:
                    df_bag.append(i)
                    frames = df_bag
            newmet = pd.concat(frames)
            newmet.index = range(0, len(newmet))
            # repalce radn data
            rad = get_nasarad(lonlat, start, end)
            newmet["radn"] = rad.ALLSKY_SFC_SW_DWN.values
            if len(newmet) != len(datecheck):
                print('date discontinuities still exisists')
            else:
                # print("met data is in the range of specified dates no discontinuities")
                rg = len(newmet.day.values) + 1
                # newmet  = pd.concat(newmet)
                mean_maxt = newmet['maxt'].mean(skipna=True, numeric_only=None)
                mean_mint = newmet['mint'].mean(skipna=True, numeric_only=None)
                AMP = round(mean_maxt - mean_mint, 2)
                tav = round(statistics.mean((mean_maxt, mean_mint)), 2)
                tile = conn.headers["Content-Disposition"].split("=")[1].split("_")[0]
                fn = conn.headers["Content-Disposition"].split("=")[1].replace("csv", 'met')
                if not filename:
                    shortenfn = generate_unique_name("Daymet") + '.met'
                else:
                    shortenfn = filename
                if not os.path.exists('weatherdata'):
                    os.makedirs('weatherdata')
                fn = shortenfn
                fname = os.path.join('weatherdata', fn)
                headers = ['year', 'day', 'radn', 'maxt', 'mint', 'rain', 'vp', 'swe']
                header_string = " ".join(headers) + "\n"
                # close and append new lines
                with open(fname, "a") as f2app:
                    f2app.writelines([f'!site: {tile}\n', f'latitude = {lonlat[1]} \n', f'longitude = {lonlat[0]}\n',
                                      f'tav ={tav}\n', f'amp ={AMP}\n'])
                    f2app.writelines([header_string])
                    f2app.writelines(['() () (MJ/m2/day) (oC) (oC) (mm) (hPa) (kg/m2)\n'])
                    # append the weather data
                    data_rows = []
                    for index, row in newmet.iterrows():
                        current_row = []
                        for header in headers:
                            current_row.append(str(row[header]))
                        current_str = " ".join(current_row) + '\n'
                        data_rows.append(current_str)

                    f2app.writelines(data_rows)

                if cleanup:
                    if os.path.isfile(os.path.join(os.getcwd(), outFname)):
                        os.remove(os.path.join(os.getcwd(), outFname))
                return fname  # fname


# dowload weather data from nasapower
def getnasa(lonlat:Union[tuple, list], start:int, end:int):
    lon = lonlat[0]
    lat = lonlat[1]
    param = ["T2M_MAX", "T2M_MIN", "ALLSKY_SFC_SW_DWN", "PRECTOTCORR", "RH2M", "WS2M"]
    pars = ",".join(param)
    rm = f'https://power.larc.nasa.gov/api/temporal/daily/point?start={start}0101&end={end}1231&latitude={lat}&longitude={lon}&community=ag&parameters={pars}&format=json&user=richard&header=true&time-standard=lst'
    data = requests.get(rm)
    dt = json.loads(data.content)
    df = pd.DataFrame(dt["properties"]['parameter'])
    return df



import numpy as np
def soilvar_perdep_cor(nlayers, soil_bottom=200, a=0.5, b=0.5):  # has potential to cythonize
    depthn = np.arange(1, nlayers + 1, 1)
    if a < 0:
        print("Target parameter can not be negative")  # a * e^(-b * x).
    elif (a > 0 and b != 0):
        ep = -b * depthn
        term1 = (a * depthn) * np.exp(ep)
        result = term1 / term1.max()
        return (result)
    elif (a == 0 and b != 0):
        ep = -b * depthn
        result = np.exp(ep) / np.exp(-b)
        return result
    elif (a == 0, b == 0):
        ans = [1] * len_layers
        return ans
    
def cal_dul(dul, bd, KS, ll15, nlayers = 10):
     DUL = dul *soilvar_perdep_cor(nlayers)
     DUL[0] = dul
     BD = bd *soilvar_perdep_cor(nlayers)
     BD[0] = bd
     LL15 =  ll15 * soilvar_perdep_cor(nlayers)
     LL15[0] = ll15
     ks = [KS] * nlayers


def cal_x(x, nlayers=10):
    xc = x * soilvar_perdep_cor(nlayers)
    xc[0] = x
   



import os
import shutil

import psutil


from apsimNGpy.settings import logger
from apsimNGpy.core.base_data import load_default_simulations
from unittest import TestCase
from os import path
from pathlib import Path
from apsimNGpy.core.core import CoreModel
from abc import ABC, abstractmethod
from os import chdir as _chdir
from os import remove
import tempfile
import joblib

wd= Path.cwd() / 'apsimNGpy_tests'
def set_wd(_wd=None):
    if _wd is None:
      WD = Path.cwd() / 'apsimNGpy_tests'
    else: WD = _wd
    WD.mkdir(exist_ok=True)
    _chdir(wd)

set_wd()
TEMPS_DIR = []
temp_dir = tempfile.TemporaryDirectory()
TEMPS_DIR.append(temp_dir)
os.chdir(temp_dir.name)

joblib.dump(TEMPS_DIR, './temps', )
from apsimNGpy.settings import SCRATCH


def get_files(pattern):
    return list(Path(SCRATCH).rglob(pattern))


def release_file_locks(_dir):
    '''it is too computationally expensive'''
    procs = psutil.process_iter()
    fd = get_files("*.apsimx")
    files = [str(i) for i in fd]
    print(files, "|||||||||||||")
    if not files:
        return None
    for proc in procs:
        try:
            for file in proc.open_files():

                if file.path in files:
                    print("Releasing file {}".format(file.path))
                    proc.terminate()
        except Exception:
            pass


class BaseTester(TestCase):
    try:
        def setUp(self):

            # Mock the model path and other attributes
            self.model_path = Path(load_default_simulations(crop='maize', simulations_object=False), )
            self.model_path2 = Path(load_default_simulations(crop='soybean', simulations_object=False), )
            self.logger = logger
            self.out_path = Path.cwd() / 'test_output.apsimx'
            self.test_ap_sim = CoreModel(model=self.model_path)

            self.out = path.realpath('test_save_output.apsimx')

        def tearDown(self):
            if path.exists(self.out):
                remove(self.out)

        @classmethod
        def setUpClass(cls):

            cls.temp_dir = temp_dir

        @classmethod
        def _tearDownClass(cls):

            try:

                cls.temp_dir.cleanup()
                shutil.rmtree(SCRATCH)
                print('Clean up completed...')
            except PermissionError:
                base = Path(wd)
                p = list(base.rglob('*.apsimx')) + list(base.rglob('*.db')) + list(base.rglob('*.db-shm'))
                for i in p:
                    try:
                        os.remove(i)
                    except PermissionError:
                        print(f'Could not remove `{i}`')

                ...
    finally:
        for temp_dir in TEMPS_DIR:
            print(temp_dir)
            ...
            # temp_dir.cleanup()



from apsimNGpy.core.pythonet_config import load_pythonnet
from Models.Soils import Soil, Physical, SoilCrop, Organic, Solute, Chemical


def soil_components(component):
    _comp = component.lower()
    comps = {'organic': Organic,
             'physical': Physical,
             'soilcrop': SoilCrop,
             'solute': Solute,
             'chemical': Chemical,
             }
    return comps[_comp]


# physical child parameters
physical_parameters = {'BD', "DUL", "Sand", 'Clay', 'Silt', "SAT", 'AirDry', 'LL15', 'SW', 'KS', 'Rocks', "Depth"}
physicals = {key: Physical for key in physical_parameters}
# organic child parameters
organic_parameters = {'Carbon', "FBiom", "FInert", 'FOM', "Depth", 'SoilCNRatio'}
organics = {key: Organic for key in organic_parameters}



import doctest
from pathlib import Path
from pprint import pprint

from apsimNGpy.core.core import CoreModel, Models

NODES = dict(manager=Models.Manager, simulation=Models.Core.Simulation,
             plant=Models.PMF.Plant,
             clock=Models.Clock,
             weather=Models.Climate.Weather,
             soil=Models.Soils.Soil,
             physical=Models.Soils.Physical,
             replacements=Models.Core.Folder)


class Inspector(CoreModel):
    """
    Inspector class for APSIMNGPY modules. It inherits from the CoreModel class and
    therefore has access to a repertoire of methods from that class.

    This implies that you can still run the model and modify parameters as needed.

    Example:
        >>> from apsimNGpy.core.inspector import Inspector
        >>> from apsimNGpy.core.base_data import load_default_simulations
        >>> path_model = load_default_simulations(crop='Maize', simulations_object=False)
        >>> model = Inspector(path_model, set_wd=Path.home())
    """

    def __init__(self, model, out_path=None, set_wd=None, **kwargs):
        super().__init__(model, out_path, set_wd, **kwargs)

    @property
    def in_simulations(self):
        return self.find_simulations()

    def inspect(self, of_class):
        of_class = NODES[of_class]
        descendants = self.Simulations.FindAllDescendants[of_class]()
        return [i.FullPath for i in descendants]

    @property
    def path_to_managers(self):
        """
        Returns a list of node paths in the class Manager.

        @return: A list of node paths in the class Manager.

        Example:
            >>> from apsimNGpy.core.base_data import load_default_simulations
            >>> model_path = load_default_simulations(crop="Maize", simulations_object=False)
            >>> model = Inspector(model_path)
            >>> pm = model.path_to_managers
            >>> pprint(pm)
            ['.Simulations.Simulation.Field.Sow using a variable rule',
             '.Simulations.Simulation.Field.Fertilise at sowing',
             '.Simulations.Simulation.Field.Harvest']
        """
        return self.inspect(of_class='manager')

    @property
    def path_to_soil(self):
        """
        Returns a list of node paths in the class Soil.

        @return: A list of node paths in the class Soil.

        Example:
            >>> from apsimNGpy.core.base_data import load_default_simulations
            >>> model_path = load_default_simulations(crop="Maize", simulations_object=False)
            >>> model = Inspector(model_path)
            >>> ps = model.path_to_soil
            >>> print(ps)
            ['.Simulations.Simulation.Field.Soil']
        """
        return self.inspect(of_class='soil')

    @property
    def path_to_plants(self):
        """
        Returns a list of node paths in the class Plants.

        @return: A list of node paths in the class Plants.

        Example:
            >>> from apsimNGpy.core.base_data import load_default_simulations
            >>> model_path = load_default_simulations(crop="Maize", simulations_object=False)
            >>> model = Inspector(model_path)
            >>> pp = model.path_to_plants
            >>> print(pp)
            ['.Simulations.Simulation.Field.Maize']
        """
        return self.inspect(of_class='plant')

    @property
    def path_to_clock(self):
        """
        Returns a list of node paths in the class Clock.

        @return: A list of node paths in the class Clock.

        Example:
            >>> from apsimNGpy.core.base_data import load_default_simulations
            >>> model_path = load_default_simulations(crop="Maize", simulations_object=False)
            >>> model = Inspector(model_path)
            >>> pc = model.path_to_clock
            >>> print(pc)
            ['.Simulations.Simulation.Clock']
        """
        return self.inspect(of_class='clock')

    @property
    def path_to_weather(self):
        """
        RReturns a list of node path in the class Weather
        @return:
        """
        return self.inspect(of_class='weather')

    @property
    def path_to_simulation(self):
        """
        RReturns a list of node path in the class Simulation
        @return:
        """
        return self.inspect(of_class='simulation')

    def get_manager_ids(self, full_path: bool = True, verbose=False) -> list[str]:

        managers = self.Simulations.FindAllDescendants[Models.Manager]()
        if full_path:
            list_managers = [i.FullPath for i in managers]
        else:
            list_managers = [i.Name for i in managers]
        if verbose:
            pprint(list_managers, indent=4)
        return list_managers

    def get_manager_parameters(self, full_path, verbose=False) -> dict:
        params = {}
        manager = self.Simulations.FindByPath(full_path)
        for i in range(len(manager.Value.Parameters)):
            _param = manager.Value.Parameters[i].Key
            _values = manager.Value.Parameters[i].Value
            params[_param] = _values
        if verbose:
            pprint(params, indent=4)
        return params

    def report_ids(self):
        datastorage = self.Simulations.FindAllDescendants[Models.Report]()
        print(datastorage)


if __name__ == '__main__':
    import doctest
    from apsimNGpy.core.base_data import load_default_simulations, load_default_sensitivity_model

    mod = load_default_sensitivity_model(method='sobol')
    sob = mod.Simulations.FindInScope[Models.Sobol]()

    model = load_default_simulations(crop='Maize')


    def set_sensitivity_factor(name, path, LowerBound, UpperBound):
        parameter = Models.Sensitivity.Parameter()
        parameter.Name = name
        parameter.LowerBound = LowerBound
        parameter.UpperBound = UpperBound
        parameter.Path = path
        return parameter


    def create_sensitivity_model(method='Sobol', **kwargs):
        ses_model = model.find_model(method)
        parameter = set_sensitivity_factor(name="CN2", path="Field.Soil.SoilWater.CN2Bare", LowerBound=70.0,
                                           UpperBound=85.0)
        parameter2 = set_sensitivity_factor(name="CN2", path="Field.Soil.SoilWater.SummerCona", LowerBound=4.0,
                                           UpperBound=9)

        sens = ses_model()


        model.add_model(model_type=sens, adoptive_parent=Models.Core.Simulations)
        sens = model.Simulations.FindInScope[ses_model](method)
        sens.set_TableName("Report")
        sens.set_AggregationVariableName('[Clock].Today')
        sens.Parameters.Add(parameter2)
        sens.Parameters.Add(parameter)
        model.move_model(model_type=Models.Core.Simulation, new_parent_type=ses_model)
        model.save()
        return model


    sob.get_BaseSimulation
    # doctest.testmod()
    for i in dir(sob):
        if 'get' in i:
            print(i)
            att = getattr(sob, i)
            if callable(att):
                try:
                    print(att())
                except Exception as e:
                    ...
            else:
                print(att)
        else:
            print(i, '__')
    mo = create_sensitivity_model()






import os.path
import platform
from pathlib import Path
from subprocess import *
from typing import Dict, Any, Union
import contextlib
import pandas as pd
from functools import lru_cache
from apsimNGpy.core.config import get_apsim_bin_path
from apsimNGpy.settings import logger
from apsimNGpy.core_utils.utils import timer
import contextlib
from apsimNGpy.settings import *
apsim_bin_path = Path(get_apsim_bin_path())

# Determine executable based on OS
if platform.system() == "Windows":
    APSIM_EXEC = apsim_bin_path / "Models.exe"
else:  # Linux or macOS
    APSIM_EXEC = apsim_bin_path / "Models"


def get_apsim_version(verbose:bool=False):
    """ Display version information of the apsim model currently in the apsimNGpy config environment.

    ``verbose``: (bool) Prints the version information ``instantly``

    Example::

            apsim_version = get_apsim_version()

    """
    cmd = [APSIM_EXEC, '--version']
    if verbose:
        cmd.append("--verbose")
    res = Popen(cmd, stdout=PIPE, stderr=PIPE, text=True)
    res.wait()
    return res.stdout.readlines()[0].strip()


def upgrade_apsim_file(file: str, verbose:bool=True):
    """
    Upgrade a file to the latest version of the .apsimx file format without running the file.

    Parameters
    ---------------
    ``file``: file to be upgraded to the newest version

    ``verbose``: Write detailed messages to stdout when a conversion starts/finishes.

    ``return``
       The latest version of the .apsimx file with the same name as the input file

    Example::

        from apsimNGpy.core.base_data import load_default_simulations
        filep =load_default_simulations(simulations_object= False)# this is just an example perhaps you need to pass a lower verion file because this one is extracted from thecurrent model as the excutor
        upgrade_file =upgrade_apsim_file(filep, verbose=False)

    """
    file = str(file)
    assert os.path.isfile(file) and file.endswith(".apsimx"), f"{file} does not exists a supported apsim file"
    cmd = [APSIM_EXEC, file, '--upgrade']
    if verbose:
        cmd.append('--verbose')
    res = Popen(cmd, stdout=PIPE, stderr=PIPE, text=True)
    outp, err = res.communicate()
    if err:
        print(err)
    if verbose:
        print(outp)
    if res.returncode == 0:
        return file


def run_model_externally(model: Union[Path, str], verbose: bool = False, to_csv:bool=False) -> Popen[str]:
    """
    Runs an ``APSIM`` model externally, ensuring cross-platform compatibility.

    Although ``APSIM`` models can be run internally, compatibility issues across different APSIM versions—
    particularly with compiling manager scripts—led to the introduction of this method.

    ``model``: (str) Path to the ``APSIM`` model file or a filename pattern.

    ``verbose``: (bool) If ``True``, prints stdout output during execution.

    ``to_csv``: (bool) If ``True``, write the results to a CSV file in the same directory.

    ``returns``: A subprocess.Popen object.

    Example::

        result =run_model_externally("path/to/model.apsimx", verbose=True, to_csv=True)
        from apsimNGpy.core.base_data import load_default_simulations
        path_to_model = load_default_simulations(crop ='maize', simulations_object =False)
        pop_obj = run_model_externally(path_to_model, verbose=False)
        pop_obj1 = run_model_externally(path_to_model, verbose=True)# when verbose is true, will print the time taken
    """

    apsim_file = model  # Define the APSIMX file path
    cmd = [str(APSIM_EXEC), str(apsim_file), '--verbose']

    if to_csv:
        cmd.append('--csv')

    with contextlib.ExitStack() as stack:
        result = stack.enter_context(Popen(cmd, stdout=PIPE, stderr=PIPE, text=True))

        try:
            out, err = result.communicate()

            if err:
                logger.error(err.strip())
            if verbose:
                logger.info(f"{repr(out.strip())}, output from {apsim_file}")

            return result
        finally:
            if result.poll() is None:  # Ensures the process is properly killed
                result.kill()


from pathlib import Path
from typing import Union, List

@lru_cache(maxsize=20)
def get_matching_files(dir_path: Union[str, Path], pattern: str, recursive: bool = False) -> List[Path]:
    """
    Search for files matching a given pattern in the specified directory.

    Args:
        ``dir_path`` (Union[str, Path]): The directory path to search in.
        ``pattern`` (str): The filename pattern to match (e.g., "*.apsimx").
        ``recursive`` (bool): If True, search recursively; otherwise, search only in the top-level directory.

    Returns:
        List[Path]: A ``list`` of matching Path objects.

    Raises:
        ``ValueError:`` If no matching files are found.
    """
    global_path = Path(dir_path)
    if recursive:
        matching_files = list(global_path.rglob(pattern))
    else:
        matching_files = list(global_path.glob(pattern))

    if not matching_files:
        raise FileNotFoundError(f"No files found that matched pattern: {pattern} in directory: {dir_path}")

    return matching_files


def collect_csv_by_model_path(model_path) -> dict[Any, Any]:
    """Collects the data from the simulated model after run
    """
    ab_path = Path(os.path.abspath(model_path))

    dirname = ab_path.parent
    name = ab_path.stem

    csv_files = list(dirname.glob(f"*{name}*.csv"))
    if csv_files:
        stems = [s.stem.split('.')[-1] for s in csv_files]
        report_paths = {k: v for k, v in zip(stems, csv_files)}

    else:
        report_paths = {}
    return report_paths

def collect_csv_from_dir(dir_path, pattern, recursive=False) -> (pd.DataFrame):
    """Collects the csf=v files in a directory using a pattern, usually the pattern resembling the one of the simulations used to generate those csv files
    ``dir_path``: (str) path where to look for csv files
    ``recursive``: (bool) whether to recursively search through the directory defaults to false:
    ``pattern``:(str) pattern of the apsim files that produced the csv files through simulations

    returns
        a generator object with pandas data frames

    Example::

         mock_data = Path.home() / 'mock_data' # this a mock directory substitute accordingly
         df1= list(collect_csv_from_dir(mock_data, '*.apsimx', recursive=True)) # collects all csf file produced by apsimx recursively
         df2= list(collect_csv_from_dir(mock_data, '*.apsimx',  recursive=False)) # collects all csf file produced by apsimx only in the specified directory directory


    """
    global_path = Path(dir_path)
    if recursive:
        matching_apsimx_patterns = global_path.rglob(pattern)
    else:
        matching_apsimx_patterns = global_path.glob(pattern)
    if matching_apsimx_patterns:
        for file_path in matching_apsimx_patterns:

            # Strip the '.apsimx' extension and prepare to match CSV files
            base_name = str(file_path.stem)

            # Search for CSV files in the same directory as the original file
            for csv_file in global_path.rglob(f"*{base_name}*.csv"):
                file = csv_file.with_suffix('.csv')
                if file.is_file():
                    df = pd.read_csv(file)
                    if isinstance(df, pd.DataFrame) and not df.empty:
                        yield df
                    else:  # notify potential issue
                        logger.warning(f"{base_name} is not a csv file or itis empty")


def run_from_dir(dir_path, pattern, verbose=False,
                 recursive=False,  # set to false because the data collector is also set to false
                 write_tocsv=True) -> [pd.DataFrame]:
    """
       This function acts as a wrapper around the ``APSIM`` command line recursive tool, automating
       the execution of APSIM simulations on all files matching a given pattern in a specified
       directory. It facilitates running simulations recursively across directories and outputs
       the results for each file are stored to a csv file in the same directory as the file'.

       What this function does is that it makes it easy to retrieve the simulated files, returning a generator that
       yields data frames

       :Parameters:
       __________________
       ``dir_path``: (str or Path, required). The path to the directory where the
           simulation files are located.
       ``pattern``: (str, required): The file pattern to match for simulation files
           (e.g., "*.apsimx").
       ``recursive``: (bool, optional):  Recursively search through subdirectories for files
           matching the file specification.
       ``write_tocsv``: (bool, optional): specify whether to write the
           simulation results to a csv. if true, the exported csv files bear the same name as the input apsimx file name
           with suffix reportname.csv. if it is ``False``,
          - if ``verbose``, the progress is printed as the elapsed time and the successfully saved csv

       ``returns``
        -- a ``generator`` that yields data frames knitted by pandas


       Example::

            mock_data = Path.home() / 'mock_data'  # As an example, let's mock some data; move the APSIM files to this directory before running
            mock_data.mkdir(parents=True, exist_ok=True)

            from apsimNGpy.core.base_data import load_default_simulations
            path_to_model = load_default_simulations(crop='maize', simulations_object=False)  # Get base model

            ap = path_to_model.replicate_file(k=10, path=mock_data) if not list(mock_data.rglob("*.apsimx")) else None

            df = run_from_dir(str(mock_data), pattern="*.apsimx", verbose=True, recursive=True)  # All files that match the pattern


       """
    dir_path = str(dir_path)

    dir_patern = f"{dir_path}/{pattern}"
    base_cmd = [str(APSIM_EXEC), str(dir_patern)]
    if recursive:
        base_cmd.append('--recursive')
    if verbose:
        base_cmd.append('--verbose')
    if write_tocsv:
        base_cmd.append('--csv')
    ran_ok = False
    with  contextlib.ExitStack() as stack:
        process = stack.enter_context(Popen(base_cmd, stdout=PIPE, stderr=PIPE,
                                            text=True))

        try:
            logger.info('waiting for APSIM simulations to complete')
            out, st_err = process.communicate()
            if verbose:
                logger.info(f'{out.strip()}')
            ran_ok = True
        finally:
            if process.poll() is None:
                process.kill()

    if write_tocsv and ran_ok:
        logger.info(f"Loading data into memory.")
        out = collect_csv_from_dir(dir_path, pattern)
        return out

@lru_cache(maxsize=None)
def apsim_executable(path, *args):
    base  = [str(APSIM_EXEC), str(path), '--verbose', '--csv']
    return base

def run_p(path):
    run(apsim_executable(path))


if __name__ == "__main__":
    import doctest
# doctest.testmod()



import os
import shutil
from pathlib import Path
from typing import Union

import apsimNGpy
import inspect
from apsimNGpy.core.apsim import ApsimModel
from apsimNGpy.core.core import CoreModel
from apsimNGpy.core import runner
from apsimNGpy.core import (config, base_data, apsim, load_model, structure)
from apsimNGpy.manager import soilmanager, weathermanager
from apsimNGpy.validation import evaluator

modules = list((config, base_data, apsim))
SENDTO = Path.cwd().parent.parent / 'docs/source'
SENDTO.mkdir(parents=True, exist_ok=True)
SENDTO2 = Path.cwd().parent.parent.parent/'apsimNGpy-documentations/doc'
SENDTO2.mkdir(parents=True, exist_ok=True)


def sort_modules(modules):
    """
    Sorts a list of modules alphabetically by their name.

    Args:
        modules (list): A list of module objects.

    Returns:
        list: A list of modules sorted by their names.
    """
    names = {i.__name__: i for i in modules}  # Create a dictionary mapping names to modules
    sorted_names = sorted(names.keys())  # Sort module names alphabetically
    return [names[name] for name in sorted_names]  # Return modules in sorted order


def docs(modules: Union[list, object], output_file: Union[str, Path] = "api.rst", skip_undocumented: bool = True,
         main_package="apsimNGpy"):
    """
    Extracts the docstrings of all public functions and classes in a module and writes them to an .rst file.

    Args:
        modules (list[module] or module): The Python modules or classes to extract documentation from.
        output_file (str): The path to the output .rst file.
        main_package (str): The name of the main package for writing the main title
    """
    if not isinstance(modules, list):
        modules = [modules]
    module_set = list(set(modules))
    modules = sort_modules(module_set)
    for i in modules:
        print(i)

    with open(output_file, "w", encoding="utf-8") as file:
        title = ''
        file.write(f"{main_package}: API Reference\n")
        file.write("-" * (len(main_package) + 15) + "\n\n")
        for module in modules:
            module_name = module.__name__
            file.write(f"{module.__name__} \n")
            file.write("-" * (len(module.__name__) + 15) + "\n\n")
            module_name = module.__name__ if hasattr(module, '__name__') else module.__class__.__name__
            class_name = 'Class' if hasattr(module, '__class__') else ''
            if inspect.isclass(module):
                sig = inspect.signature(module)
                file.write(f".. function:: {module.__module__}.{module.__name__}{sig}\n\n")
                doc = module.__doc__ or "No documentation available."
                file.write(f"   {doc.strip()}\n\n")
            if inspect.isfunction(module):
                if module_name not in module.__module__:
                    continue
                if not module.__name__.startswith('_'):
                    sig = inspect.signature(module)
                    file.write(f".. function:: {module.__module__}.{module.__name__}{sig}\n\n")
                    doc = module.__doc__ or "No documentation available."
                    file.write(f"   {doc.strip()}\n\n")

            # if inspect.getdoc(module) is not None and hasattr(module, '__class__'):
            #     file.write(f"{inspect.getdoc(module)}\n")

            # Extract public functions
            for name, func in inspect.getmembers(module, inspect.isfunction):
                is_class = inspect.isclass(module)
                if is_class and func.__module__ != module.__module__:
                    continue
                if 'apsimNGpy' not in func.__module__.split("."):
                    continue
                if name.startswith("_"):
                    continue  # Skip private functions
                if not inspect.getdoc(func) and skip_undocumented:  # Skip undocumented funcs
                    continue
                sig = inspect.signature(func)
                if is_class:
                    file.write(f".. function:: {module.__module__}.{module.__name__}.{name}{sig}\n\n")
                else:
                    file.write(f".. function:: {func.__module__}.{name}{sig}\n\n")

                doc = func.__doc__ or "No documentation available."
                file.write(f"   {doc.strip()}\n\n")

            # Extract public classes and their public methods
            for name, cls in inspect.getmembers(module, inspect.isclass):
                cls_name = cls.__name__

                if 'apsimNGpy' not in cls.__module__.split("."):
                    continue
                if name.startswith("_"):
                    continue  # Skip private classes
                if not inspect.getdoc(cls) and skip_undocumented:
                    continue  # Skip writing undocumented
                file.write(f".. class:: {cls.__module__}{name}\n\n")
                class_doc = cls.__doc__ or "No documentation available."
                file.write(f"   {class_doc.strip()}\n\n")

                for method_name, method in inspect.getmembers(cls, inspect.isfunction):
                    if method_name.startswith("_"):
                        continue  # Skip private methods
                    if not inspect.getdoc(method) and skip_undocumented:
                        continue
                    if method.__module__ != cls.__module__:
                        continue
                    sig = inspect.signature(method)
                    file.write(f"   .. method::{method.__module__}.{cls_name}.{method_name}{sig}\n\n")
                    method_doc = method.__doc__ or "No documentation available."
                    file.write(f"      {method_doc.strip()}\n\n")

    print(f"✅ API reference documentation written to: {output_file}")


def dc():
    ...


if __name__ == '__main__':
    from apsimNGpy.core_utils import database_utils
    from apsimNGpy.parallel import process
    from apsimNGpy.optimizer.one_obj import ContinuousVariableProblem

    runs = [
        runner.collect_csv_by_model_path,
        runner.run_model_externally,
        runner.run_from_dir,
        runner.upgrade_apsim_file, ]



    from apsimNGpy.core import core

    docs([apsim.ApsimModel, ContinuousVariableProblem, process, database_utils, core.CoreModel, core.ModelTools, evaluator, runner, base_data, weathermanager, soilmanager, load_model],
         output_file="api.rst")

    rsts = list(Path.cwd().rglob("*pi.rst")) #+ list(Path.cwd().rglob("*conf.py"))
    if SENDTO2.exists():
        sd2 = SENDTO2
        for rst in rsts:
            # turn off index
            if str(rst).endswith("dex.rst") or str(rst).endswith("inspection.rst"):
                continue
            file_name = SENDTO2.joinpath(rst.name)
            if file_name.exists():
                os.remove(file_name)
            copy = shutil.copy2(rst, file_name)

            if copy.stem =='api':

                ... #os.startfile(copy)
            print(copy)
    else:
        print(f"{SENDTO2} not present")

    for rst in rsts:
        if str(rst).endswith("index.rst") or str(rst).endswith("inspection.rst"):
            continue
        file_name = SENDTO.joinpath(rst.name)
        if file_name.exists():
            os.remove(file_name)

        copy = shutil.copy2(rst, file_name)
        print(copy)

    # sphinx-build -b html docs/source docs/build/html



import os

import pandas as pd

from apsimNGpy.parallel.process import run_apsimx_files_in_parallel, read_result_in_parallel
from apsimNGpy.core_utils.utils import collect_runfiles
from pathlib import Path

hd = Path.home()/'scart'
hd.mkdir(parents=True, exist_ok=True)
os.chdir(hd)
from apsimNGpy.core.base_data import load_default_simulations
from apsimNGpy.core_utils.utils import make_apsimx_clones

# let's duplicate some files

# let's copy two files here
maize= load_default_simulations(crop='maize', simulations_object= False)

ap = make_apsimx_clones(maize, 20)


def fnn(x):
    return x


# that is it all these files are now in the dir_path
files = collect_runfiles(path2files=hd, pattern=["*.apsimx"])  # change path as needed

if __name__ == "__main__":
    # copy all examples to our workig dir_path
    # ex = apsim_example.get_all_examples()
    xp =[i for i in run_apsimx_files_in_parallel(ap, ncores=10, use_threads=False)]
    # files is an iterable or a generator
    # ncores is the number of process or threads to use and use_threads will determine wthere we use threads or not
    # returns nothing
    rpath = os.path.realpath(hd.joinpath('apsimx_cloned'))
    # we can now read the results using the .db path as follow
    files_db = [i for i in collect_runfiles(path2files=rpath, pattern=["*.db"])]
    dat = read_result_in_parallel(files_db, ncores=2, use_threads=True, report_name='report')
    # return a generator object
    dl = list(dat)
    # df = pd.concat(dat)
    print(dl)



from concurrent.futures import as_completed

import numpy as np
from typing import Tuple, Any
from apsimNGpy.core.apsim import ApsimModel
from apsimNGpy.core.base_data import LoadExampleFiles
from pathlib import Path

from apsimNGpy.core_utils.utils import select_process
from apsimNGpy.weather import daymet_bylocation_nocsv, daymet_bylocation
from apsimNGpy.manager.soilmanager import DownloadsurgoSoiltables, OrganiseSoilProfile
from apsimNGpy.core_utils.spatial import create_fishnet1, create_apsimx_sim_files, generate_random_points
from tqdm import tqdm

wd = Path.home()
from os.path import basename

maize = LoadExampleFiles(wd).get_maize

lon = -91.620369, 43.034534


def simulate_single_point(model: Any, location: Tuple[float, float], report, read_from_string=True, start=1990,
                          end=2020,
                          soil_series: str = 'domtcp', **kwargs):
    """
    Run a simulation of a given crop.
     model: Union[str, Simulations],
     location: longitude and latitude to run from, previously lonlat
     soil_series: str
     kwargs:
        copy: bool = False, out_path: str = None, read_from_string=True,

        soil_series: str = 'domtcp', thickness: int = 20, bottomdepth: int = 200,

        thickness_values: list = None, run_all_soils: bool = False

        report_name: str specifies the report or table name in the simulation, for which to read the reasults

        replace_weather: Set this boolean to true to download and replace the weather data based on the specified location.

        replace_soil: Set this boolean to true to download and replace the soil data using the given location details.

        mgt_practices: Provide a list of management decissions

    """
    thi = [150, 150, 200, 200, 200, 250, 300, 300, 400, 500]
    th = kwargs.get("thickness_values", thi)  # in case it is not supplied, we take thi
    simulator_model = ApsimModel(
        model, copy=kwargs.get('copy'), read_from_string=read_from_string, lonlat=location, thickness_values=th)
    # if replace weather
    sim_name = simulator_model.extract_simulation_name
    if kwargs.get('replace_weather', False):
        wname = model.strip('.apsimx') + '_w.met'
        wf = daymet_bylocation_nocsv(location, start, end, filename=wname)
        simulator_model.replace_met_file(wf, sim_name)
    # replace soil is true
    if kwargs.get("replace_soil", False):
        table = DownloadsurgoSoiltables(location)
        sp = OrganiseSoilProfile(table, thickness=20, thickness_values=th)
        sp = sp.cal_missingFromSurgo()
        simulator_model.replace_downloaded_soils(sp, sim_name)
    # if replace management practices
    if kwargs.get("mgt_practices"):
        simulator_model.update_mgt(kwargs.get('mgt_practices'), sim_name)
    simulator_model.run(report_name=report)
    return simulator_model.results


def simulate_single_from_shape_file(wd, shape_file, model: Any, resolution, report, read_from_string=True, start=1990,
                                    end=2020,
                                    soil_series: str = 'domtcp', **kwargs):
    arr = create_fishnet1(shape_file, lon_step=resolution, lat_step=resolution, ncores=3)
    """
    Run a simulation of a given crop.
     model: Union[str, Simulations],
     location: longitude and latitude to run from, previously lonlat
     soil_series: str
     wd: pathlike stirng
     kwargs:
        copy: bool = False, out_path: str = None, read_from_string=True,

        soil_series: str = 'domtcp', thickness: int = 20, bottomdepth: int = 200,

        thickness_values: list = None, run_all_soils: bool = False

        report_name: str specifies the report or table name in the simulation, for which to read the reasults

        replace_weather: Set this boolean to true to download and replace the weather data based on the specified location.

        replace_soil: Set this boolean to true to download and replace the soil data using the given location details.

        mgt_practices: Provide a list of management decissions
        
        ncores: set the number of cores
        use_thread: set true to run in parallel processing
        random_grid_points: bolean. set this true to sample specified poitns per grid of your choice
        num_points: int,  set number of points

    """
    use_thread, ncores, num_points = kwargs.get('use_thread', True), kwargs.get('ncores', 3), kwargs.get('num_points',
                                                                                                         2)
    if kwargs.get('random_grid_points'):
        arr = generate_random_points(shape_file, resolution, ncores, num_points)
    else:
        arr = create_fishnet1(shape_file, lon_step=resolution, lat_step=resolution, ncores=kwargs.get('ncores', 4))
    df = create_apsimx_sim_files(wd, model, arr)
    thi = [150, 150, 200, 200, 200, 250, 300, 300, 400, 500]
    th = kwargs.get("thickness_values", thi)  # in case it is not supplied, we take thi

    def worker(row):
        ID = row['ID']
        model = row['file_name']
        simulator_model = ApsimModel(
            model, copy=kwargs.get('copy'), read_from_string=read_from_string, lonlat=None, thickness_values=th)
        sim_names = simulator_model.extract_simulation_name
        location = row['location']

        if kwargs.get('replace_weather', False):
            wname = model.strip('.apsimx') + '_w.met'
            wf = daymet_bylocation_nocsv(location, start, end, filename=wname)
            simulator_model.replace_met_file(wf, sim_names)

        if kwargs.get("replace_soil", False):
            table = DownloadsurgoSoiltables(location)
            sp = OrganiseSoilProfile(table, thickness=20, thickness_values=th)
            sp = sp.cal_missingFromSurgo()
            simulator_model.replace_downloaded_soils(sp, sim_names)

        if kwargs.get("mgt_practices"):
            simulator_model.update_mgt(kwargs.get('mgt_practices'), sim_names)
        simulator_model.run(report_name=report)
        return simulator_model.results

    with select_process(use_thread, ncores) as tpool:
        futures = {tpool.submit(worker, df.loc[df['ID'] == i].squeeze()): i for i in df['ID']}
        progress = tqdm(total=len(futures), position=0, leave=True,
                        bar_format=f'Running:' '{percentage:3.0f}% completed')
        # Iterate over the futures as they complete
        for future in as_completed(futures):
            yield future.result()
            progress.update(1)
        progress.close()


md = {"Name": 'PostharvestillageMaize', 'Fraction': 0.001
      }
if __name__ == '__main__':
    pp = simulate_single_point(maize, lon, replace_weather=True, replace_soil=True, mgt_practices=md, report='MaizeR')
    shp = r'D:\ACPd\Bear creek simulations\bearcreek_shape\bearcreek.shp'
    wd = r'C:\Users\rmagala'
    lp = simulate_single_from_shape_file(wd, shp, maize, 1200, report='MaizeR', random_grid_points=True, replace_soil=True, replace_weather=True)
    lip = list(lp)



# this an evolving module acording to github issue No#

import typer
import requests

app = typer.Typer()

BASE_URL = "http://127.0.0.1:8000"

@app.command()
def version():
    """Get APSIM version from the server."""
    response = requests.get(f"{BASE_URL}/version")
    if response.ok:
        typer.echo(f"APSIM Version: {response.json()['apsim_version']}")
    else:
        typer.echo("Failed to fetch version.")

@app.command()
def get_path():
    """Get current APSIM bin path."""
    response = requests.get(f"{BASE_URL}/get-path")
    if response.ok:
        typer.echo(f"Current APSIM bin path: {response.json()['current_bin_path']}")
    else:
        typer.echo("Failed to fetch bin path.")

@app.command()
def set_path(path: str):
    """Set the APSIM bin path."""
    response = requests.get(f"{BASE_URL}/set-path", params={"update": path})
    if response.ok:
        typer.echo(response.json()["message"])
    else:
        typer.echo(f"Error: {response.status_code} - {response.text}")

@app.command()
def auto_detect():
    """Auto detect APSIM bin path."""
    response = requests.get(f"{BASE_URL}/auto-detect")
    if response.ok:
        typer.echo(f"Detected: {response.json()['detected_path']}")
    else:
        typer.echo("No bin path detected.")
@app.command()
def load(name: str) -> str:
    """Send a request to load the model on the server."""
    response = requests.post(f"{BASE_URL}/load", json={"name": name})
    response.raise_for_status()
    print('loaded')
    return response.json()["message"]
@app.command()
def run() -> str:
    """Trigger model execution on the server."""
    response = requests.post(f"{BASE_URL}/run")
    response.raise_for_status()
    return response.json()["message"]
@app.command()
def inspect(_model:str):
    response = requests.get(f"{BASE_URL}/inspect", json={"_model": _model})
    if response.ok:
        typer.echo(f"Detected: {response.json()["model_inspect"]}")
    else:
        typer.echo("No model inspected.")

if __name__ == "__main__":
    app()



import logging
import os
import shutil
from contextlib import contextmanager
from pathlib import Path

from apsimNGpy.core.config import auto_detect_apsim_bin_path
from apsimNGpy.core.config import get_apsim_bin_path
from apsimNGpy.settings import logger, MSG

try:
    from apsimNGpy.core.core import CoreModel
    from apsimNGpy.core.apsim import ApsimModel
    from apsimNGpy.core.base_data import load_default_simulations
except (ModuleNotFoundError, ImportError):
    ...
current_path = os.path.dirname(os.path.abspath(__file__))
# Set up basic configuration for logging
logging.basicConfig(format='%(asctime)s :: %(message)s', level=logging.INFO)

user_id = current_path

auto = auto_detect_apsim_bin_path() or os.getenv('APSIM')


def test():
    # test auto detect;
    if auto:
        logging.info(f"apsim path detected at: {auto}")
    # test pythonnet
    if get_apsim_bin_path():
        logging.info('Using apsim path at: {}'.format(get_apsim_bin_path()))
    else:
        logger.debug(MSG)
    from pathlib import Path
    from time import perf_counter
    # Model = FileFormat.ReadFromFile[Models.Core.Simulations](model, None, False)
    os.chdir(Path.home())
    from apsimNGpy.core.base_data import load_default_simulations

    model = load_default_simulations(crop='maize')
    logging.info(f"testing simulator time\n ______________________")
    for _ in range(2):
        a = perf_counter()
        # model.RevertCheckpoint()

        model.run(report_name='Report')
        # print(model.results.mean(numeric_only=True))

        # print(model.results.mean(numeric_only=True))
        msg = f"{perf_counter() - a} seconds, taken"
        logging.info(msg=msg, )


from apsimNGpy.experiment.main import Experiment


@contextmanager
def test_experiment(use_tread=False):
    path = Path.home().joinpath('scratchT')
    path.mkdir(exist_ok=True)

    try:
        # Import the model from APSIM and get the model path
        model_path = load_default_simulations('maize').path
        logging.info("Testing experiment module\n ______________________")

        # Set up the experiment
        FactorialExperiment = Experiment(
            database_name='test.db',
            datastorage='test.db',
            tag='th',
            base_file=model_path,
            wd=path,
            use_thread=use_tread,
            skip_completed=False,
            verbose=False,
            test=False,
            n_core=6,
            reports={'Report'}
        )

        # Add factors
        FactorialExperiment.add_factor(
            parameter='Carbon',
            param_values=[1.4, 0.2, 0.4, 0.9, 2.4, 0.6, 0.8],
            factor_type='soils',
            soil_node='Organic'
        )
        FactorialExperiment.add_factor(
            parameter='FBiom',
            param_values=[0.045, .4, 2.4, 0.8, 0.7],
            factor_type='soils',
            soil_node='Organic'
        )

        # Clear the database and start the experiment
        FactorialExperiment.clear_data_base()
        FactorialExperiment.start_experiment()

        # Yield control back to the block inside the `with` statement
        yield FactorialExperiment

        # Log successful test after the experiment block
        logging.info("Experiment module test successful\n____________________________________")

    except Exception as e:
        # Log any exceptions encountered during the experiment
        try:
            logging.exception(f"Experiment module test failed due to exception: {repr(e)}")
        except PermissionError as e:
            ...

    finally:
        # Clean up by removing the temporary dir_path
        try:
            shutil.rmtree(path)
            logging.info("files removed successfully")
        except PermissionError as e:
            ...


# Running the test
if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    test()
    with test_experiment(use_tread=True) as experiment:
        sim_data = experiment.get_simulated_data()[0]
        mn = sim_data.groupby(['FBiom'])['Yield'].mean()
        logging.info(f"Simulation results (mean yield by FBiom):\n{mn}")



import argparse
from apsimNGpy.cli.set_ups import ColoredHelpFormatter
from apsimNGpy.cli.commands import change_mgt
parser = argparse.ArgumentParser(description='\033[91mRun a simulation of a given crop.\033[0m',
                                 formatter_class=ColoredHelpFormatter)

# Add arguments
parser.add_argument('-m', '--model', type=str, required=False,
                    help='Path  or name of the APSIM model file if path it should ends with .apsimx'
                         'defaults to maize from the default '
                         'simulations ', default='Maize')
parser.add_argument('-o', '--out', type=str, required=False, help='Out path for apsim file')
parser.add_argument('-mg', '--management', type=str, required=False, help=f'{change_mgt.__doc__}')
parser.add_argument('-p', '--preview', type=str, required=False, choices=('yes', 'no'), default='no',
                    help='Preview or start model in GUI')
parser.add_argument('-sv', '--save', type=str, required=False,
                    help='Out path for apsim file to save after making changes')
parser.add_argument('-t', '--table', type=str, required=False, help='Report table name. '
                                                                    'Defaults to "Report"')
parser.add_argument('-w', '--met_file', type=str, required=False, help="Path to the weather data file.")
parser.add_argument('-i', '--inspect', type=str, required=False, help=f"inspect your model to get the model paths.")
parser.add_argument('-sim', '--simulation', type=str, required=False, help='Name of the APSIM simulation to run')
parser.add_argument('-ws', '--wd', type=str, required=False, help='Working directory')
parser.add_argument('-l', '--lonlat', type=str, required=False, help='longitude and Latitude (comma-separated) '
                                                                     'for fetching weather data.')
parser.add_argument('-sm', '--save_model', type=str, required=False, help='File name for saving output data.')
parser.add_argument('-s', '--aggfunc', type=str, required=False, default='mean',
                    help='Statistical summary function (e.g., mean, median). Defaults to "mean". Useful for quick diagnostics,'
                         'but real data is stored in csf file')
parser.add_argument('-og', '--organic', type=str,
                    required=False,
                    help="Replace any soil data through a soil organic parameters and path specification"
                         " e.g, 'node_path=.Simulations.Simulation.Field.Soil, Carbon=[2.2]'")
parser.add_argument('-ph', '--physical', type=str,
                    required=False,
                    help="Replace any soil data through a soil physical parameters and path specification"
                         " e.g, 'node_path=.Simulations.Simulation.Field.Soil, BD=[1.2]'")
parser.add_argument('-ch', '--chemical', type=str,
                    required=False,
                    help="Replace any soil data through a soil chemical parameters and path specification"
                         " e.g, 'node_path=.Simulations.Simulation.Field.Soil, NH4=[2.2]'")

parser.add_argument('-fw', '--get_web_data', type=str, required=False, choices=['both', 's', 'w', 'no'],
                    default='no', help=' if lonlat arguments is provided, get soil or weather data'
                                       ' w flag get weather and s flag gets soil data while both will download both')
parser.add_argument('-e', '--experiment', type=str, required=False,
                    help="set experiment by providing the specification in the form of a string."
                         "Different specifications are seperated by a  full colon."
                         "Edits like changing management practices or soils _variables will be made before creating the experiment ")



from dataclasses import dataclass
from apsimNGpy.core.runner import run_from_dir, get_matching_files
import pandas as pd
import polars as pl
from apsimNGpy.core_utils.database_utils import (read_db_table, get_db_table_names,custom_remove_file,
                                                 clear_all_tables, pl_save_todb)
from os.path import basename
from sqlalchemy import create_engine
from apsimNGpy.core_utils.utils import timer
from pathlib import Path
from datetime import datetime
import shutil, os, sys
from apsimNGpy.parallel.process import custom_parallel
from apsimNGpy.core.apsim import ApsimModel
DateTime =  datetime.now().strftime("%Y-%m-%d-%H%M%S")
@dataclass
class BatchRunner:
    path: str
    pattern: str = '*.apsimx'
    export_csv: bool = False
    recursive : bool = True
    verbose: bool = False
    results: pd.DataFrame = None
    def run(self, results = True):
        df = run_from_dir(self.path, pattern=self.pattern, write_tocsv=self.export_csv,
                          verbose=self.verbose,recursive=self.recursive)
        self.results = list(df)
    def edit_run_model(self):
        find_files = []

        try:
            work_space = Path(self.path)/f'{DateTime}'
            # freshen work_space_exists:
            work_space.mkdir(parents=True, exist_ok=True)
            model = ApsimModel('Maize').replicate_file(k=200, path=work_space)
            try:
               find_files = get_matching_files(work_space, self.pattern)
            except FileNotFoundError:
                find_files = []
            if find_files:
                custom_parallel(custom_remove_file, find_files, progress_message='removing old files')
                print('success')
        finally:
            tl = custom_parallel(custom_remove_file, find_files, void = False, progress_message='Removing old files')
            list(tl)
            shutil.rmtree(work_space)
            ex = os.path.exists(work_space)
            print(ex)

        ...
    @timer
    def export_all_todb(self,database:str):
            database= database if database.endswith('.db') else database + '.db'
            pattern  =self.pattern.replace('.apsimx','.csv')
            csvs = get_matching_files(self.path, pattern, self.recursive)
            if  csvs:
                base_names = [Path(i).stem.split(".")[-1] for i in csvs]
                print(base_names)
                dfs = [pl.read_csv(c) for c in csvs if c is not None]
                datas= [(k, v) for k, v in zip(base_names, dfs)]

                list(custom_parallel(pl_save_todb, datas, database, 'replace', void =True, progress_message="saving data to {}".format(database)))

                ...
            else:
                print('No results')

if __name__ == '__main__':
    from pathlib import Path
    mock_data = Path.home() / 'mock_data'  # As an example let's mock some data move the apsim files to this directory before runnning
    mock_data.mkdir(parents=True, exist_ok=True)
    from apsimNGpy.core.base_data import load_default_simulations
    path_to_model = load_default_simulations(crop='maize', simulations_object=True)  # get base model
    ap = path_to_model.replicate_file(k=10, path=mock_data) if not list(mock_data.rglob("*.apsimx")) else None
    # df = run_from_dir(str(mock_data), pattern="*.apsimx", verbose=True,
    #                        recursive=True)  # all files that matches that pattern
    Batch = BatchRunner(mock_data, pattern='*.apsimx', export_csv=True, recursive=False, verbose=True)
    Batch.export_all_todb('dd')
    #Batch.edit_run_model()




import logging


class DualLogger:
    def __init__(self, name='apsimNGpy', log_file='apsimNGpy_sim.log', console_level=logging.DEBUG,
                 file_level=logging.INFO):
        """
        Initialize the DualLogger.

        Parameters:
        -----------
        name : str
            The name of the logger.
        log_file : str
            The filename where logs will be saved.
        console_level : int
            The logging level for the console output.
        file_level : int
            The logging level for the file output.
        """
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.DEBUG)

        # Create console handler and set level
        console_handler = logging.StreamHandler()
        console_handler.setLevel(console_level)

        # Create file handler and set level
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(file_level)

        # Create formatter and set it for both handlers
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        console_handler.setFormatter(formatter)
        file_handler.setFormatter(formatter)

        # Add handlers to the logger
        self.logger.addHandler(console_handler)
        self.logger.addHandler(file_handler)

    def __call__(self, level, message):
        """
        Log a message at the given level.

        Parameters:
        -----------
        level : str
            The level at which to log the message ('debug', 'info', 'warning', 'error', 'critical').
        message : str
            The message to log.
        """
        if level == 'debug':
            self.logger.debug(message)
        elif level == 'info':
            self.logger.info(message)
        elif level == 'warning':
            self.logger.warning(message)
        elif level == 'error':
            self.logger.error(message)
        elif level == 'critical':
            self.logger.critical(message)
        else:
            raise ValueError(f"Invalid log level: {level}")



import os.path
import unittest

from apsimNGpy.tests.base_test import BaseTester

from apsimNGpy.manager import weathermanager

LONLAT = 35.582520, 0.347596  # location is kampala
AMES_LONLAT = -93.0, 42.03534


class test_weatherManager(BaseTester):
    def test_get_weather(self):
        met = weathermanager.get_weather(LONLAT, source='nasa')
        self.assertIsInstance(met, str)
        self.assertTrue(os.path.exists(met))
        met = weathermanager.get_weather(AMES_LONLAT, source='daymet')
        self.assertIsInstance(met, str)
        self.assertTrue(os.path.exists(met))
        self.logger.info('get_weather called successfully')

    def test_is_within_usa_mainland(self):
        type_bool = weathermanager._is_within_USA_mainland(LONLAT)
        assert isinstance(type_bool, bool)
        self.logger.info('is_within_USA_mainland ran successfully')


if __name__ == '__main__':
    unittest.main()




import logging
import os
import traceback

import matplotlib.pyplot as plt
from pygments.lexers import nit
import copy
from config import *

logger = DualLogger()
cap = ORR
info = 'info'
error = 'error'
SoilTable = 'ORR'


def _filter_Comp(_sdata, comp_name):
    return _sdata[_sdata['componentname'] == comp_name].copy()


series = 'soilseriesname1'

Thickness = [200, 100, 200, 200, 200, 250, 300, 300, 350, 250]
comp_names = ['Osco', 'Sable', 'Muscatune', 'Drummer', 'Denny', 'Edgington',
              'Ipava', 'Elburn', 'Buckhart']

getSoils = pd.read_excel(cap, sheet_name='Soil')
_soils = getSoils.dropna(subset=['plotid']).copy()


def cleanLoadSoilDf(cap_xlsx_path, soil_table_db_name, filter_from: int = 2012):
    Cp_names = COMPS[soil_table_db_name]
    """ a dat frame of soils form a cap experimental sites
     soil_df: pandas dataframe containing soils properties or results from the experiment
     filter_from:int - the year to filter out defaults to 2011
     """
    # load plot data
    #########################################################
    try:
        Sdata_tables = read_db_table(wdi.path(name='NWREC_db.DB_MP'), report_name=soil_table_db_name)

        SPdata = Sdata_tables.copy(deep=True)
        cn = SPdata.componentname.unique().tolist()

        STables = {comps: _filter_Comp(SPdata, comps) for comps in cn}

        soilProfiles = {}
        for Comps in STables.keys():
            try:
                des = OrganizeAPSIMsoil_profile(STables[Comps], thickness=20, state='Michigan').cal_missingFromSurgo()
                soilProfiles[Comps] = des
            except ValueError as e:
                pass
        newValidComps = soilProfiles.keys()

        plots = pd.read_excel(cap_xlsx_path, sheet_name='Plot Identifiers')
        plots = plots[(plots['soil_data'] == 'yes') & (plots['agro_data'] == 'yes') & (
                plots['tillage'] == 'TIL2')]
        if plots.empty:
            raise ValueError('plots empty+++++++++++++++++')
        else:
            print(plots)
        plt_soil = plots.drop_duplicates(['plotid', 'soilseriesname1'])
        print(plt_soil['soilseriesname1'].unique())
        new_sdf = plt_soil.loc[plt_soil['soilseriesname1'].isin(newValidComps)]

        crop_data = ['2011crop', '2012crop', '2013crop', '2014crop', '2015crop']
        rep = {'Corn': 'Maize'}
        plt_soil = plt_soil
        plt_soil.replace(rep, inplace=True)
        cps = plt_soil[crop_data].copy()
        cropSequences = [create_crop_sequence(cps, n) for n in cps.index.to_list()]
        print(plt_soil.shape)
        plt_soil['cropSequences'] = cropSequences
        # load soils data sheet name is soils
        ###########################################################################
        soil_dff = pd.read_excel(cap_xlsx_path, sheet_name='Soil')
        soilDF = soil_dff.copy(deep=True)
        _soiD = soilDF.drop([0, 1], axis=0).copy()
        soils = _soiD[_soiD['year'] == filter_from]
        print(_soiD.year.unique())
        print(soils.shape)
        soc = 'SOIL13'
        ph = 'SOIL11'
        bd = 'SOIL01'
        nitrates = 'SOIL15'
        dul = 'SOIL31'
        l15 = 'SOIL32'
        # new_sdf['plotid'] = new_sdf['plotid'].astype('float')
        # soils['plotid'] =  soils['plotid'].astype('float')
        grad_x = new_sdf.merge(soils, on='plotid', how='inner')
        print('merge_succeeded')
        grad_x.rename(columns={soc: 'Carbon', ph: 'PH', bd: 'BD', nitrates: 'N03', dul: 'DUL', l15: 'LL15'},
                      inplace=True)
        kpSoils = soil_dff.rename(columns={soc: 'Carbon', ph: 'PH', bd: 'BD', nitrates: 'N03', dul: 'DUL', l15: 'LL15'},
                                  inplace=False).copy()
        cpd = new_sdf[crop_data].copy()
        new_sdf['cropSequences'] = [create_crop_sequence(cps, n) for n in cpd.index.to_list()]
        dtm = grad_x, kpSoils, soilProfiles, new_sdf

        return dtm
    except Exception as e:
        logger('info', f" {traceback.format_exc()}, {repr(e)}")


def _ratio(ar):
    return ar / ar[0]


grad_x, kpSoils, soilProfiles, new_sdf = cleanLoadSoilDf(XLs[SoilTable], soil_table_db_name=SoilTable)


def organize(df, plotid):
    try:
        if df.shape[0] == 0:
            raise ValueError('No data to organize, \n')
        """
        Organize the soils by plotid
        :param df:
        :param plotid:
        :return:
        """
        _df = df[df['plotid'] == plotid].copy()

        cp_id = _df['soilseriesname1'].iloc[0]
        till = _df['tillage'].iloc[0]

        _df.drop_duplicates(['depth'], inplace=True)
        _df = _df.copy()
        _df = _df[(_df['depth'] == '10 to 20') | (_df['depth'] == '0 to 10')].copy()
        _df = _df[['DUL', "BD", 'LL15', 'PH', 'Carbon']].copy()

        variable = _df
        variable.replace('M', -999, inplace=True)
        variable = variable.astype(float)
        variable = variable.mean(numeric_only=True).to_frame().T.copy()
        data_2 = soilProfiles.get(cp_id)

        data_ = copy.deepcopy(data_2)
        bd = np.array(data_[0]['BD'])
        dul = np.array(data_[0]['DUL'])
        l15 = np.array(data_[0]['LL15'])
        _carbon = np.array(data_[1]['Carbon'])
        ad = np.array([0.5, 0.8, 0.8, 1, 1, 1, 1, 1, 1, 1])
        assert len(ad) == 10, 'Length not equal to target'

        def _check_naN(param):
            varT = variable[param].to_list()
            if varT[0] == -999:
                # print(param, 'will not be imputed because it is empty or missing')
                return None
            else:
                print(f"{param}: passed")
                return True

        if _check_naN('BD'):
            data_[0]['BD'] = _ratio(bd) * np.array(variable.get('BD')).tolist()
        if _check_naN('DUL'):
            data_[0]['DUL'] = _ratio(dul) * np.array(variable.get('DUL')).tolist()
        if _check_naN('LL15'):
            data_[0]['LL15'] = _ratio(l15) * np.array(variable.get('LL15')).tolist()
        if _check_naN('Carbon'):
            data_[1]['Carbon'] = _ratio(_carbon) * np.array(variable.get('Carbon') / 10).tolist()
        # adjust SAT according to Bulk-density
        data_[0]['SAT'] = 1 - (data_[0]['BD'] / 2.65)
        saT, duL = data_[0]['SAT'].copy(), data_[0]['DUL'].copy()
        if till == 'TIL1':
            print('No tillage detected')
            data_[0].loc[0, 'BD'] = data_[0].loc[0, 'BD'] * 0.97
            data_[1].loc[0, 'FBiom'] = 0.4
            data_[1].loc[0, 'FInert'] = 0.8
            data_[1].loc[1, 'FInert'] = 0.9
        print('adjusting SAT & DUL')

        def adjustSatDul(sat_, dul_):
            for enum, (s, d) in enumerate(zip(sat_, dul_)):
                # first check if they are equal
                # if d is greater than s, then by what value?, we need this value to add it to 0.02
                #  to be certain all the time that dul is less than s we subtract the summed value
                if d >= s:

                    diff = d - s
                    if diff == 0:
                        dul_[enum] = d - 0.02
                    else:
                        duL[enum] = d - (diff + 0.02)


                else:
                    dul_[enum] = d
            return dul

        duL = adjustSatDul(saT, duL)
        data_[0]['DUL'] = list(duL)

        data_[0]['AirDry'] = list(np.array(data_[0]['LL15']) * ad)
        return data_
    except (IndexError, TypeError) as ide:
        pass


# od = organize(grad_x, plotid= 1071.0)

calSoilProfiles = {}
for i in new_sdf['plotid']:
    print('filling missing data')
    try:

        xp = organize(grad_x, i)
        calSoilProfiles[i] = xp
    # vc= xp[0].SAT
    except Exception as e:
        logger(error, f"Full traceback: {traceback.format_exc()}")
        pass

from apsimNGpy.core.apsim import ApsimModel

run = True
PLots = [1052.0,
         1081.0,
         1092.0,
         2012.0,
         2052.0,
         2071.0,
         3011.0,
         3062.0,
         3101.0,
         4031.0,
         4062.0,
         4082.0]
if __name__ == '__main__' and run:
    apsi_m = wdi.path(name='apsim/baseC.apsimx')
    directory = wdi.path(name='NWREC_edited')
    wdi.mkdir('NWREC_edited')
    model = ApsimModel(apsi_m)
    plot_path = {}
    Crops = 'Maize'


    # NREW N fertilizer was 246
    def _mgt(crop, depth=350, n_amount=220):

        if 'Soybean' in crop:
            n_amount = 170
        sowManager = {"Name": 'MaizeNitrogenManager', 'Amount': n_amount}
        return [{"Name": 'Simple Rotation', 'Crops': crop},
                {'Name': 'Tillage', 'Depth': depth, 'AppplyTillageOption': 'no', 'Fraction': 0.65}, sowManager]


    predicted = []
    pYield = []
    those = []
    for i in PLots:
        pp = [i]
    for soils_s in PLots:  # calSoilProfiles.keys():#calSoilProfiles.keys():
        # th = str(soils_s)
        # if th.count('31')> 0:
        try:
            croP = new_sdf[new_sdf['plotid'] == soils_s]['cropSequences'].iloc[0]

            out_path = os.path.join(directory, f"{soils_s}.apsimx")
            _model = ApsimModel(apsi_m, thickness_values=Thickness)
            sdp = copy.deepcopy(calSoilProfiles[soils_s])
            _model.replace_downloaded_soils(sdp, simulation_names=_model.extract_simulation_name)
            # _model.replace_met_file(mETS[SoilTable], simulations=model.extract_simulation_name)
            _model.update_mgt(_mgt(croP))
            mf = _model.show_met_file_in_simulation()
            # print(croP)
            if croP.count('Maize') > 4:
                print(croP, '++_________')
                _model.save_edited_file(outpath=out_path)
                # _model = ApsimModel(out_path)
                _model.run(dynamicReportSpec(croP, Maize=['Annual', 'MaizeR']))
                pc, Yield = _model.results
                pc['plotid'] = soils_s
                # pc['year'] = pc['an_Year'].astype(float)
                pc['CropSequences'] = croP
                Yield['year'] = Yield['Year'].astype(float)
                Yield['plotid'] = soils_s
                Yield['CropSequences'] = croP
                plot_path[soils_s] = out_path
                those.append(soils_s)
                predicted.append(pc)
                # os.startfile(_model.path)
                print(soils_s, ":::", croP)

                pYield.append(Yield)
        except IndexError:
            pass
    import joblib

    predicted_ = pd.concat(predicted)
    Yield = pd.concat(pYield)
    joblib.load(os.path.join(directory, 'orr_edited_files.joblib'))
    apprent_plots = plot_path.keys()
    soc = kpSoils[kpSoils['plotid'].isin(apprent_plots)].copy()
    soc1 = soc.drop_duplicates(subset=['plotid', 'depth', 'year'])
    soc2 = soc1[soc1['Carbon'] != 'M'].copy()
    soc3 = soc2[['plotid', 'depth', 'year', 'Carbon', 'BD']].copy()
    soc4 = soc3[(soc3['depth'] == '10 to 20') | (soc3['depth'] == '0 to 10')].copy()

    soc4['Carbon'] = soc4['Carbon'].astype('float')
    soc4 = soc4[soc4 != 'M'].dropna().copy()
    soc4['BD'] = soc4['BD'].astype('float')
    grp_agg = {"Carbon": 'mean', 'BD': 'mean'}
    soc5 = soc4.groupby(['plotid', 'year']).agg(grp_agg)
    soc5.reset_index(inplace=True)
    soc5['depth'] = 20
    df = soc5.copy()
    df['soc'] = df['depth'] * df['BD'] * (df['Carbon'] / 10)
    predicted_.rename(columns={'an_Year': 'year'}, inplace=True)
    df.to_csv(wdi.path(name='observed_soc.csv'))
    data = df.merge(predicted_, on=['plotid', 'year'], how='inner')
    data['socx'] = data['SOC1'] / (data['BD'] * data['depth'])
    data['Carbon'] = data['Carbon'] / 10
    data.to_csv(wdi.path(name='carbon.csv'))
    dat = data.copy()
    # data = dat[dat['year']==2014.0].copy()
    val = validate(actual=data.Carbon, predicted=data.socx)
    ev_all = val.evaluate_all(verbose=True)
    X = data[['Carbon']].values
    y = data['socx'].values
    plot_reg_fit(X, y, data=data)

    C_yield = load_AGRON(Yield, XLs[SoilTable])
    mY = validate(actual=C_yield.obs / 1000, predicted=C_yield.Maize_Yield / 1000)
    mY.evaluate_all(verbose=True)
    from apsimNGpy.visual.visual import quick_plot

    # print(C_yield)
    plot_po(xx=data.year, predicted=data.socx, observed=data.Carbon, anotate_with=None)
    data.to_csv(wdi.path(name='ccs.csv'))

    print(those)



import json, os, sys
from os.path import join as opj
import numpy as np
import numpy
import requests
import xmltodict
import pandas as pd
import time
from numpy import array as npar
import sys
from scipy import interpolate
import traceback
from datetime import datetime
import datetime
import copy
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt

THICKNESS = [150, 150, 200, 200, 200, 250, 300, 300, 400, 500]


def DownloadsurgoSoiltables(lonlat, select_componentname=None, summarytable=False):
    '''
    Downloads SSURGO soil tables
    
    parameters
    ------------------
    ``lon``: longitude

    ``lat``: latitude

    ``select_componentname``: any componet name within the map unit e.g 'Clarion'. the default is None that mean sa ll the soil componets intersecting a given locationw il be returned
      if specified only that soil component table will be returned. in case it is not found the dominant componet will be returned with a caveat meassage.
        use select_componentname = 'domtcp' to return the dorminant component.

    ``summarytable``: prints the component names, their percentages
    '''

    total_steps = 3
    # lat = "37.54189"
    # lon = "-120.96683"
    lonLat = "{0} {1}".format(lonlat[0], lonlat[1])
    url = "https://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx"
    # headers = {'content-child': 'application/soap+xml'}
    headers = {'content-child': 'text/xml'}
    body = """<?xml version="1.0" encoding="utf-8"?>
              <soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:sdm="http://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx">
       <soap:Header/>
       <soap:Body>
          <sdm:RunQuery>
             <sdm:Query>SELECT co.cokey as cokey, ch.chkey as chkey, comppct_r as prcent, compkind as compkind_series, wsatiated_r as wat_r,partdensity as pd, dbthirdbar_h as bb, musym as musymbol, compname as componentname, muname as muname, slope_r, slope_h as slope, hzname, hzdept_r as topdepth, hzdepb_r as bottomdepth, awc_r as PAW, ksat_l as KSAT,
                        claytotal_r as clay, silttotal_r as silt, sandtotal_r as sand, om_r as OM, iacornsr as CSR, dbthirdbar_r as BD, wfifteenbar_r as L15, wthirdbar_h as DUL, ph1to1h2o_r as pH, ksat_r as sat_hidric_cond,
                        (dbthirdbar_r-wthirdbar_r)/100 as bd FROM sacatalog sc
                        FULL OUTER JOIN legend lg  ON sc.areasymbol=lg.areasymbol
                        FULL OUTER JOIN mapunit mu ON lg.lkey=mu.lkey
                        FULL OUTER JOIN component co ON mu.mukey=co.mukey
                        FULL OUTER JOIN chorizon ch ON co.cokey=ch.cokey
                        FULL OUTER JOIN chtexturegrp ctg ON ch.chkey=ctg.chkey
                        FULL OUTER JOIN chtexture ct ON ctg.chtgkey=ct.chtgkey
                        FULL OUTER JOIN copmgrp pmg ON co.cokey=pmg.cokey
                        FULL OUTER JOIN corestrictions rt ON co.cokey=rt.cokey
                        WHERE mu.mukey IN (SELECT * from SDA_Get_Mukey_from_intersection_with_WktWgs84('point(""" + lonLat + """)')) 
                        
                        AND sc.areasymbol != 'US' 
                        order by co.cokey, ch.chkey, prcent, topdepth, bottomdepth, muname
            </sdm:Query>
          </sdm:RunQuery>
       </soap:Body>
    </soap:Envelope>"""

    response = requests.post(url, data=body, headers=headers, timeout=140)
    # Put query results in dictionary format
    my_dict = xmltodict.parse(response.content)

    # Convert from dictionary to dataframe format

    soil_df = pd.DataFrame.from_dict(
        my_dict['soap:Envelope']['soap:Body']['RunQueryResponse']['RunQueryResult']['diffgr:diffgram']['NewDataSet'][
            'Table'])
    if summarytable:
        df = soil_df.drop_duplicates(subset=['componentname'])
        summarytable = df[["componentname", 'prcent', 'chkey']]
        print("summary of the returned soil tables \n")
        print(summarytable)
    # selec the dominat componet
    dom_component = soil_df[soil_df.prcent == soil_df.prcent.max()]
    # select by component name
    if select_componentname in soil_df.componentname.unique():
        componentdf = soil_df[soil_df.componentname == select_componentname]
        return componentdf
    elif select_componentname == 'domtcp':
        return dom_component
        # print("the following{0} soil components were found". format(list(soil_df.componentname.unique())))
    elif select_componentname == None:
        # print("the following{0} soil components were found". format(list(soil_df.componentname.unique())))
        return soil_df
    elif select_componentname != 'domtcp' and select_componentname not in soil_df.componentname.unique() or select_componentname != None:
        print(
            f'Ooops! we realised that your component request: {select_componentname} does not exists at the specified location. We have returned the dorminant component name')
        return dom_component


# test the function
# aa= DownloadsurgoSoiltables([-90.72704709, 40.93103233],'Osco', summarytable = True)


##Making APSIM soil profile starts here=============================
len_layers = 10
a = 1.35
b = 1.4


# create a variable profile constructor
def soilvar_perdep_cor(nlayers, soil_bottom=200, a=0.5, b=0.5):  # has potential to cythonize
    depthn = np.arange(1, nlayers + 1, 1)
    if a < 0:
        print("Target parameter can not be negative")  # a * e^(-b * x).
    elif (a > 0 and b != 0):
        ep = -b * depthn
        term1 = (a * depthn) * np.exp(ep)
        result = term1 / term1.max()
        return (result)
    elif (a == 0 and b != 0):
        ep = -b * depthn
        result = np.exp(ep) / np.exp(-b)
        return result
    elif (a == 0, b == 0):
        ans = [1] * len_layers
        return ans


def set_depth(depththickness):
    """
  parameters
  depththickness (array):  an array specifying the thicknness for each layer
  nlayers (int); number of layers just to remind you that you have to consider them
  ------
  return
bottom depth and top depth in a turple
  """
    # thickness  = np.tile(thickness, 10
    thickness_array = np.array(depththickness)
    bottomdepth = np.cumsum(thickness_array)  # bottom depth should nothave zero
    top_depth = bottomdepth - thickness_array
    return bottomdepth, top_depth


distparms = {'a': 0, 'b': 0.2}


class OrganizeAPSIMsoil_profile:
    # Iinitiate the soil object and covert all into numpy array and change them to floating values
    def __init__(self, sdf, thickness, thickness_values=None, bottomdepth=200):
        """_summary_

        Args:
            sdf (pandas data frame): soil table downloaded from SSURGO_
            thickness double: _the thickness of the soil depth e.g 20cm_
            bottomdepth (int, optional): _description_. Defaults to 200.
            thickness_values (list or None) optional if provided extrapolation will be based on those vlue and should be the same length as the existing profile depth
         """
        sdf1 = sdf.drop_duplicates(subset=["topdepth"])
        surgodf = sdf1.sort_values('topdepth', ascending=True)
        self.clay = npar(surgodf.clay).astype(np.float16)
        self.sand = npar(surgodf.sand).astype(np.float16)
        self.silt = npar(surgodf.silt).astype(np.float16)
        self.OM = npar(surgodf.OM).astype(np.float16)
        self.topdepth = npar(surgodf.topdepth).astype(np.float16)
        self.bottomdepth = npar(surgodf.bottomdepth).astype(np.float16)
        self.BD = npar(surgodf.bb).astype(np.float16)
        self.DUL = npar(surgodf.DUL).astype(np.float64)
        self.L15 = npar(surgodf.L15).astype(np.float64)
        self.PH = npar(surgodf.pH).astype(np.float64)
        self.PAW = npar(surgodf.PAW).astype(np.float16)
        self.saturatedhudraulic_conductivity = npar(surgodf.sat_hidric_cond).astype(np.float16)
        self.KSAT = npar(surgodf.KSAT).astype(np.float16)
        self.particledensity = npar(surgodf.pd).astype(np.float16)
        self.muname = surgodf.muname
        self.musymbol = surgodf.musymbol
        self.cokey = surgodf.cokey
        self.slope = surgodf.slope_r
        self.componentname = surgodf.componentname
        self.Nlayers = bottomdepth / thickness
        self.thickness = thickness * 10
        self.wat_r = surgodf.wat_r
        self.chkey = surgodf.chkey
        self.componentpercent = surgodf.prcent
        self.newtopdepth = np.arange(0, bottomdepth, thickness)
        self.newbottomdepth = np.arange(thickness, bottomdepth + thickness, thickness)
        # trial
        self.thickness_values = np.array(thickness_values)

    # create a function that creates a variable profile of the provide _variables
    @staticmethod
    def set_depth(depththickness):
        """
        parameters
        depththickness (array):  an array specifying the thicknness for each layer
        nlayers (int); number of layers just to remind you that you have to consider them
        ------
        return
      bottom depth and top depth in a turple
        """
        # thickness  = np.tile(thickness, 10
        thickness_array = np.array(depththickness)
        bottomdepth = np.cumsum(thickness_array)  # bottom depth should nothave zero
        top_depth = bottomdepth - thickness_array
        return bottomdepth, top_depth

    def variable_profile(self, y, kind='linear'):  # has potential to cythonize
        # use the lower depth boundary because divinding by zeros is nearly impossible
        x = self.bottomdepth
        # replace x with the anticipated value of the the new soil depth
        nlayers = self.Nlayers
        xranges = x.max() - x.min()
        newthickness = xranges / nlayers
        # new depth variable for interpolation
        xnew = np.arange(x.min(), x.max(), newthickness)
        # create an interpolation function
        yinterpo = interpolate.interp1d(x, y, kind=kind, assume_sorted=False, fill_value='extrapolate')
        if isinstance(self.thickness_values, str):  # just put a strign to evaluate to false i will fix it later
            tv = np.array(self.thickness_values)
            tv = tv.astype('float64')
            xnew, top_dep = OrganizeAPSIMsoil_profile.set_depth(tv)
        apsimvar = yinterpo(xnew)
        return apsimvar

    def decreasing_exponential_function(self, x, a, b):  # has potential to cythonize
        """
          Compute the decreasing exponential function y = a * e^(-b * x).

          Parameters:
              ``x`` (array-like): Input values.
              ``a`` (float): Amplitude or scaling factor.
              ``b`` (float): Exponential rate.

          ``Returns:``
              numpy.ndarray: The computed decreasing exponential values.
          """
        func = a * np.exp(-b * x)
        return func

    def optimize_exponetial_data(self, x_data, y_data, initial_guess=[0.5, 0.5],
                                 bounds=([0.1, 0.01], [np.inf, np.inf])):  # defaults for carbon

        best_fit_params, _ = curve_fit(self.decreasing_exponential_function, x_data, y_data, p0=initial_guess,
                                       bounds=bounds)
        a_fit, b_fit = best_fit_params
        predicted = self.decreasing_exponential_function(x_data, a_fit, b_fit)
        return predicted

    def cal_satfromBD(self):  # has potential to cythonize
        if any(elem is None for elem in self.particledensity):
            pd = 2.65
            bd = npar(self.BD)
            sat = ((2.65 - bd) / 2.65) - 0.02
            sat = self.variable_profile(sat)
            return sat
        else:
            pd = self.particledensity
            bd = npar(self.BD)
            sat = ((2.65 - bd) / pd) - 0.02
            sat = self.variable_profile(sat)
            return sat
            # Calculate DUL from sand OM and clay

    def cal_dulFromsand_clay_OM(self):  # has potential to cythonize
        clay = self.clay * 0.01
        sand = self.sand * 0.01
        om = self.OM * 0.01
        ret1 = -0.251 * sand + 0.195 * clay + 0.011 * om + (0.006) * sand * om - 0.027 * clay * om + 0.452 * (
                    sand * clay) + 0.299
        dul = ret1 + 1.283 * np.float_power(ret1, 2) - 0.374 * ret1 - 0.015
        dulc = self.variable_profile(dul)
        return dulc

    def cal_l15Fromsand_clay_OM(self):  # has potential to cythonize
        clay = self.clay * 0.01
        sand = self.sand * 0.01
        om = self.OM * 0.01
        ret1 = -0.024 * sand + (
                    0.487 * clay) + 0.006 * om + 0.005 * sand * om + 0.013 * clay * om + 0.068 * sand * clay + 0.031
        ret2 = ret1 + 0.14 * ret1 - 0.02
        l151 = self.variable_profile(ret2)
        return l151

    def calculateSATfromwat_r(self):  # has potential to cythonize
        if all(elem is None for elem in npar(self.particledensity)) == False:
            wat = self.wat_r
            return self.variable_profile(wat)

    def cal_KS(self):  # has potential to cythonize
        ks = self.saturatedhudraulic_conductivity * npar([1e-06]) * (60 * 60 * 24) * 1000
        n = int(self.Nlayers)
        KS = np.full(shape=n, fill_value=ks[1], dtype=np.float64)
        return KS

    def cal_Carbon(self):  # has potential to cythonize
        ## Brady and Weil (2016)
        carbonn = self.variable_profile(self.OM) / 1.72
        # if carbonans >=3.5:
        #   carbonn = carbonans *0.3
        # else:
        #   carbonn = carbonans
        # print(carbonn)

        xdata = self.newbottomdepth
        try:
            predicted = self.optimize_exponetial_data(xdata, carbonn)

            return predicted
        except Exception as e:
            print("error occured while optimizing soil carbon to the new depth \nplease see:", repr(e))
            return carbonn

    def interpolate_clay(self):
        return self.variable_profile(self.clay)

    def interpolate_silt(self):
        return self.variable_profile(self.silt)

    def interpolate_sand(self):
        return self.variable_profile(self.sand)

    def interpolate_PH(self):
        return self.variable_profile(self.PH)

    def get_L15(self):
        if any(np.isnan(self.L15)):
            L15 = self.cal_l15Fromsand_clay_OM()
            return L15
        else:
            l1 = self.L15 * 0.01
            L15 = self.variable_profile(l1)
            return L15

    def get_DUL(self):
        if any(np.isnan(self.L15)):
            L15i = self.cal_l15Fromsand_clay_OM()

        else:
            l1 = self.L15 * 0.01
            L15i = self.variable_profile(l1)

        if any(np.isnan(self.DUL)):
            DUL = self.cal_dulFromsand_clay_OM()
            return DUL
        else:
            if not any(np.isnan(self.PAW)):
                paw = self.variable_profile(self.PAW)
                DUL = L15i + paw
                return DUL

    def get_AirDry(self):
        if any(np.isnan(self.L15)):
            air = self.cal_l15Fromsand_clay_OM()
            air[:3] = air[:3] * 0.5
            return air
        else:
            air = self.L15 * 0.01
            airl = self.variable_profile(air)
            airl[:3] = airl[:3] * 0.5
            return airl

    def getBD(self):
        return self.variable_profile(self.BD)

    def create_soilprofile(self):
        n = int(self.Nlayers)
        Depth = []
        for i in range(len(self.newbottomdepth)):
            Depth.append(str(self.newtopdepth[i]) + "-" + str(self.newbottomdepth[i]))
        Depth = Depth
        # Thickness  = [self.thickness]*self.Nlayers
        Carbon = self.cal_Carbon()
        AirDry = self.get_AirDry()
        L15 = self.get_L15()
        DUL = self.get_DUL()
        if all(elem is None for elem in npar(self.wat_r)) == False:

            SAT = self.calculateSATfromwat_r() * 0.01

            for i in range(len(SAT)):
                if SAT[i] < DUL[i]:
                    SAT[i] = DUL[i] + 0.001
            else:
                SAT[i] = SAT[i]
            BD = (1 - SAT) * 2.65
        else:
            SAT = self.cal_satfromBD()
            BD = self.getBD()
            for i in range(len(SAT)):
                if SAT[i] < DUL[i]:
                    SAT[i] = DUL[i] + 0.001
            else:
                SAT[i] = SAT[i]
            for i in range(len(SAT)):  # added it
                if SAT[i] > 0.381 and BD[i] >= 1.639:
                    SAT[i] = 0.381

        KS = self.cal_KS()
        PH = self.interpolate_PH()
        ParticleSizeClay = self.interpolate_clay()
        ParticleSizeSilt = self.interpolate_silt()
        ParticleSizeSand = self.interpolate_sand()
        df = pd.DataFrame(
            {"Depth": Depth, "Thickness": [self.thickness] * n, "BD": BD, "AirDry": AirDry, "LL15": L15, "DUL": DUL,
             "SAT": SAT, "KS": KS, "Carbon": Carbon,
             "PH": PH, "ParticleSizeClay": ParticleSizeClay, "ParticleSizeSilt": ParticleSizeSilt,
             "ParticleSizeSand": ParticleSizeSand})
        return (df)

    def exponential_function_inc_yvalue(x, a, b):
        return a * np.exp(-b * x)

    def optimize_exp_increasing_y_values(self, y_data):
        if isinstance(self.thickness_values, np.ndarray):

            xdata = np.cumsum(np.array(self.thickness_values))
        else:
            xdata = np.cumsum(np.array(self.newbottomdepth))
        xdata = xdata.astype('float64')

        # Sort the data points by increasing y-values
        sorted_indices = np.argsort(y_data)
        sorted_x_data = xdata[sorted_indices]
        sorted_y_data = y_data[sorted_indices]
        # Perform curve fitting with the sorted data
        best_fit_params, _ = curve_fit(self.exponential_function_inc_yvalue, sorted_x_data, sorted_y_data)

        # Extract the best-fit parameters (a_fit and b_fit)
        a_fit, b_fit = best_fit_params
        predcited = self.exponential_function_inc_yvalue(xdata, a_fit, b_fit)

        # Plot the original data and the fitted curve
        plt.scatter(xdata, y_data, label='Original Data')
        plt.plot(xdata, self.exponential_function_inc_yvalue(xdata, a_fit, b_fit), label='Fitted Curve', color='red')
        plt.xlabel('x')
        plt.ylabel('y')
        plt.legend()
        plt.show()
        return predcited

    def cal_missingFromSurgo(self, curveparam_a=0, curveparam_b=0.2, crops=["Wheat", "Maize", "Soybean", "Rye"],
                             metadata=None, soilwat=None, swim=None, soilorganicmatter=None):
        nlayers = int(self.Nlayers)
        # ad  = self.get_AirDry()[1]
        cropLL = self.get_AirDry()
        # Original thought
        # ad * soilvar_perdep_cor(nlayers, a = curveparam_a, b = curveparam_b)
        cropKL = 0.06 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=curveparam_b)
        cropXF = 1 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=0)
        # create a data frame for these three _variables
        dfs = pd.DataFrame({'kl': cropKL, 'll': cropLL, 'xf': cropXF})
        SoilCNRatio = np.full(shape=nlayers, fill_value=12.2, dtype=np.int64)
        FOM = 160 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=curveparam_b)
        FOMCN = np.full(shape=nlayers, fill_value=40, dtype=np.int64)
        FBiom = 0.045 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=curveparam_b)
        Fi = 0.83 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=-0.01)
        # to do later
        # try:
        #   predicted= self.optimize_exp_increasing_y_values(np.array(Fi))
        #   FInert = predicted
        # except Exception as e:
        #   print("error optiming FInert value encountered;", repr(e))
        FInert = Fi

        NO3N = 0.5 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=0.01)
        NH4N = 0.05 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=0.01)
        FInert[0] = 0.65
        FInert[1] = 0.668
        FBiom[0] = 0.0395
        FBiom[1] = 0.035
        # from above
        Carbon = self.cal_Carbon()
        PH = self.interpolate_PH()
        # vn = npar
        # PH = 6.5 * soilvar_perdep_cor(nlayers, a = curveparam_a, b = 0)
        organic = pd.DataFrame(
            {'Carbon': Carbon, 'SoilCNRatio': SoilCNRatio, 'cropLL': cropLL, 'cropKL': cropKL, 'FOM': FOM,
             'FOM.CN': FOMCN, 'FBiom': FBiom, 'FInert': FInert, 'NO3N': NO3N, 'NH4N': NH4N, 'PH': PH})
        # create a list to store the crop names
        names = []
        for i in crops:
            names.append([i + "KK", i + 'LL ', i + 'XF'])
        cropframe = []
        for i in names:
            cropframe.append(dfs.rename(columns={"kl": i[0], "ll": i[1], "xf": i[2]}))
        # print(cropframe)
        # pd.concat is faster
        cropdf = pd.concat(cropframe, join='outer', axis=1)
        physical = self.create_soilprofile()

        # create a alist
        frame = [physical, organic, cropdf, metadata]
        # All soil data frames
        resultdf = pd.concat(frame, join='outer', axis=1)
        finalsp = {'soil ': resultdf, 'crops': crops, 'metadata': metadata, 'soilwat': soilwat, 'swim': swim,
                   'soilorganicmatter ': soilorganicmatter}
        # return pd.DataFrame(finalsp)
        return frame





"""
This allows for mixed variable optimization by encoding the categorical _variables
"""
from pathlib import Path

import numpy as np

from apsimNGpy.core_utils.utils import timer

from apsimNGpy.optimizer.simple_problem import Problem, Solvers, auto_guess
from scipy.optimize import minimize, differential_evolution
import subprocess
from tqdm import tqdm
from apsimNGpy.optimizer.one_obj import AbstractProblem, SIMULATIONS, ContinuousVariableProblem, cache, VarDesc
try:
    import wrapdisc
except ModuleNotFoundError as mnf:
    print('installing wrapdisc package')
    package_name = "wrapdisc"
    command = ["pip", "install", package_name]
    # Running the command
    subprocess.run(command, check=True)

import operator
from typing import Any

import scipy.optimize

from wrapdisc import Objective
from wrapdisc.var import ChoiceVar, GridVar, QrandintVar, QuniformVar, RandintVar, UniformVar

@cache
def _variable_type(type_name: str) -> str:
    variable_types = {
        'choice': ChoiceVar,
        'grid': GridVar,
        'qrandint': QrandintVar,
        'quniform': QuniformVar,
        'randint': RandintVar,
        'uniform': UniformVar
    }
    try:
       return variable_types[type_name.lower()]
    except KeyError:
          raise ValueError(f"Invalid type '{type_name}'. Use one of: {', '.join(var_map)}")


class MixedVariableProblem(ContinuousVariableProblem):
    """
           Defines an optimization problem for continuous variables in APSIM simulations.

           This class enables the user to configure and solve optimization problems involving continuous
           control variables in APSIM models. It provides methods for setting up control variables,
           applying bounds and starting values, inserting variable values into APSIM model configurations,
           and running optimization routines using local solvers or differential evolution.

           Inherits from:
               ``ContinuousVariableProblem``

           Parameters:
               ``model (str):`` The name or path of the APSIM template file.
               .
               ``simulation (str or list, optional)``: The name(s) of the APSIM simulation(s) to target.
                                                   Defaults to all simulations.

               ``control_vars`` (list, optional): A list of VarDesc instances defining variable metadata.

               ``labels (list, optional)``: Variable labels for display and results tracking.

               ``cache_size (int):`` Maximum number of results to store in the evaluation cache.

           Attributes:
               ``model (str):`` The APSIM model template file name.
               ``simulation (str):`` Target simulation(s).
               ``controls (list):`` Defined control variables.
               ``control_vars (list):`` List of VarDesc instances for optimization.
               ``labels (list): Labels`` for variables.
               ``pbar (tqdm):`` Progress bar instance.
               ```cache (bool):`` Whether to cache evaluation results.
               ```cache_size (int):`` Size of the local cache.

           Methods:
               ``add_control(...):`` Add a new control variable to the optimization problem.
               ``bounds:`` Return the bounds for all control variables as a tuple.
               ``starting_values():`` Return the initial values for all control variables.
               ``minimize_with_local_solver(...):`` Optimize using `scipy.optimize.minimize`.
               ``optimize_with_differential_evolution(...):`` Optimize using `scipy.optimize.differential_evolution`.
               ``_open_pbar(labels, maxiter):`` Open a progress bar.
               ``_close_pbar():`` Close the progress bar.

           Example:
               >>> class Problem(ContinuousVariableProblem):
               ...     def evaluate(self, x):
               ...         return -self.run(verbose=False).results.Yield.mean()

               >>> problem = Problem(model="Maize", simulation="Sim")
               >>> problem.add_control("Manager", "Sow using a rule", "Population", int, 5, bounds=[2, 15])
               >>> result = problem.minimize_with_local_solver(method='Powell')
               >>> print(result.x_vars)
           """
    def __init__(self, model: str,
                 simulation=SIMULATIONS,
                 controls=None,
                 control_vars=None,

                 func=None,
                 cache_size=400):

        # Initialize parent classes explicitly
        AbstractProblem.__init__(self, model, simulation)
        self.model = model
        self.simulation = simulation if simulation is not SIMULATIONS else 'all'
        self.controls = controls or []
        self.control_vars = control_vars or []
        self.labels =  []
        self.cache = False
        self.cache_size = cache_size
        self.max_cache_size =cache_size
        self.pbar = None


    def add_control(
            self,
            model_type,
            model_name,
            parameter_name,
            vtype,
            start_value,
            bounds=None,
            categories=None,
            values=None,
            q=None
    ) -> "ContinuousVariableProblem":
        """
        Adds a control variable to the optimization problem. Under the hood, the variables are edited using ``edit_method``.
        So, take note of the similarities in the requested parameters

        Parameters:
            ``model_type (str)``: APSIM model type (e.g., 'Manager').
            ``model_name (str)``: Name of the model instance.
            ``parameter_name (str)``: Name of the parameter to control.
            ``vtype (type)``: Variable type class (e.g., 'choice, grid, qrandint, quniform, randint, uniform'
           ).
            ``start_value``: Initial value for the variable.
            ``bounds (tuple, optional)``: Lower and upper bounds for numeric variables.
            ``categories (list, optional)``: Category names for categorical variables.
            ``values (list, optional)``: Grid values for GridVar.
            ``q (float, optional)``: Quantization value for quantized variables.

        Returns:
            self: Enables method chaining.
        """

        self._evaluate_args(model_type, model_name, parameter_name, vtype)

        label = f"{parameter_name}"
        vtype_cls = _variable_type(vtype)

        if vtype_cls == ChoiceVar:
            _vtype = ChoiceVar(categories=categories)
        elif vtype_cls == GridVar:
            _vtype = GridVar(values)
        elif vtype_cls == QuniformVar:
            if not bounds or q is None:
                raise ValueError("QuniformVar requires bounds and q.")
            _vtype = QuniformVar(lower=bounds[0], upper=bounds[1], q=q)
        elif vtype_cls == RandintVar:
            if not bounds:
                raise ValueError("RandintVar requires bounds.")
            _vtype = RandintVar(lower=bounds[0], upper=bounds[1])
        elif vtype_cls == QrandintVar:
            if not bounds or q is None:
                raise ValueError("QrandintVar requires bounds and q.")
            _vtype = QrandintVar(lower=bounds[0], upper=bounds[1], q=q)
        else:
            raise TypeError(f"Unsupported variable type: {vtype_cls}")
        self.labels.append(label)
        self.control_vars.append(
            VarDesc(
                model_type=model_type,
                model_name=model_name,
                parameter_name=parameter_name,
                vtype=_vtype,
                label=label,
                start_value=start_value,
                bounds=bounds
            )
        )

        return self  # Enable method chaining
    def _insert_controls(self, x) -> None:

        edit = self.edit_model

        for i, varR in enumerate(self.control_vars):
            vtype = varR.vtype
            value = x[i]
            edit(
                model_type=varR.model_type,
                simulations=self.simulation,
                model_name=varR.model_name,
                cacheit=True,
                **{varR.parameter_name: value}
            )


        return x

    @property
    def _variables(self):
        var_s = []
        for index, value in enumerate(self.control_vars):
           var_s.append(value.vtype)
        return var_s

    from scipy.optimize import differential_evolution
    from tqdm import tqdm
    def _set_objective_function(self, x):
        xl = self._insert_controls(x)
        SCORE  = self.evaluate(xl)
        return SCORE

    def optimize_with_differential_evolution(self,
                                             args=(),
                                             strategy='best1bin',
                                             maxiter=1000,
                                             popsize=15,
                                             tol=0.01,
                                             mutation=(0.5, 1),
                                             recombination=0.7,
                                             rng=None,
                                             callback=None,
                                             disp=True,
                                             polish=True,
                                             init='latinhypercube',
                                             atol=0,
                                             updating='immediate',
                                             workers=1,
                                             constraints=(),
                                             x0=None,
                                             *,
                                             integrality=None,
                                             vectorized=False):
        """
        Runs differential evolution on the wrapped objective function.
        Reference: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.differential_evolution.html
        """

        try:

            self._open_pbar(labels = [l.label for l in self.control_vars])
            call_counter = {"count": 0}
            func = getattr(self, '_set_objective_function')
            def wrapped_obj(x, *args):
                call_counter["count"] += 1
                self.pbar.update(1)
                return self._set_objective_function(x)

            wrapped = Objective(
                func= wrapped_obj,
                variables=self._variables
            )

            bounds = wrapped.bounds


            # Handle initial guess
            initial_guess = self.starting_values()

            encoded_initial = wrapped.encode(initial_guess)


            # Run optimization
            result = differential_evolution(
                wrapped,
                bounds=wrapped.bounds,
                args=args,
                strategy=strategy,
                maxiter=maxiter,
                popsize=10,
                tol=tol,
                mutation=mutation,
                recombination=recombination,
                seed=0,
                disp=disp,
                polish=polish,
                init=init,
                atol=atol,
                updating='deferred',
                workers=workers,
                constraints=constraints,
                x0=encoded_initial,
                integrality=integrality,
                vectorized=vectorized
            )

            # Attach labeled solution
            decoded_solution = wrapped.decode(result.x)
            result.x_vars = dict(zip(self.labels, decoded_solution))
            self.outcomes = result
            return result

        finally:
            self.clear_cache()
            self._close_pbar()

    def minimize_with_local_solver(self, **kwargs):
        """
            Run a local optimization solver (e.g., Powell, L-BFGS-B, etc.) on given defined problem.

            This method wraps ``scipy.optimize.minimize`` and handles mixed-variable encoding internally
            using the `Objective` wrapper from ``wrapdisc``. It supports any method supported by SciPy's
            `minimize` function and uses the encoded starting values and variable bounds. This decoding implies that you can optimize categorical variable such as start dates or
            cultivar paramter with xy numerical values.

            Progress is tracked using a progress bar, and results are automatically decoded and stored
            in ``self.outcomes``.

            Parameters:
                **kwargs: Keyword arguments passed directly to `scipy.optimize.minimize`.
                          Important keys include:
                            - ``method (str)``: Optimization algorithm (e.g., 'Powell', 'L-BFGS-B').
                            - ``options (dict)``: Dictionary of solver options like maxiter, disp, etc.
        scipy.optimize.minimize provide a number of optimization algorithms see table below or for details check their website:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html#scipy.optimize.minimize

        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | Method           | Type                   | Gradient Required | Handles Bounds | Handles Constraints | Notes                                        |
        +==================+========================+===================+================+=====================+==============================================+
        | Nelder-Mead      | Local (Derivative-free)| No                | No             | No                  | Simplex algorithm                            |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | Powell           | Local (Derivative-free)| No                | Yes            | No                  | Direction set method                         |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | CG               | Local (Gradient-based) | Yes               | No             | No                  | Conjugate Gradient                           |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | BFGS             | Local (Gradient-based) | Yes               | No             | No                  | Quasi-Newton                                 |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | Newton-CG        | Local (Gradient-based) | Yes               | No             | No                  | Newton's method                              |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | L-BFGS-B         | Local (Gradient-based) | Yes               | Yes            | No                  | Limited memory BFGS                          |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | TNC              | Local (Gradient-based) | Yes               | Yes            | No                  | Truncated Newton                             |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | COBYLA           | Local (Derivative-free)| No                | No             | Yes                 | Constrained optimization by linear approx.   |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | SLSQP            | Local (Gradient-based) | Yes               | Yes            | Yes                 | Sequential Least Squares Programming         |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-constr     | Local (Gradient-based) | Yes               | Yes            | Yes                 | Trust-region constrained                     |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | dogleg           | Local (Gradient-based) | Yes               | No             | No                  | Requires Hessian                             |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-ncg        | Local (Gradient-based) | Yes               | No             | No                  | Newton-CG trust region                       |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-exact      | Local (Gradient-based) | Yes               | No             | No                  | Trust-region, exact Hessian                  |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-krylov     | Local (Gradient-based) | Yes               | No             | No                  | Trust-region, Hessian-free                   |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+

            Returns:
                result (OptimizeResult): The result of the optimization, with an additional
                                         `x_vars` attribute that provides a labeled dict of optimized
                                         control variable values.

            Raises:
                Any exceptions raised by `scipy.optimize.minimize`.

            Example:
            --------
            The following example shows how to use this method, the evaluation is very basic, but you
            can add a more advanced evaluation by adding a loss function e.g RMSE os NSE by comparing with the observed and predicted,
            and changing the control variables::

            class Problem(MixedVariableProblem):
                def __init__(self, model=None, simulation='Simulation'):
                    super().__init__(model, simulation)
                    self.simulation = simulation

                def evaluate(self, x, **kwargs):
                    # All evlauations can be defined inside here, by taking into accound the fact that the results object returns a data frame
                    # Also, you can specify the database table or report name holding the ``results``
                    return -self.run(verbose=False).results.Yield.mean() # A return is based on your objective definition, but as I said this could a ``RRMSE`` error or any other loss function

            # now we are ready to initialise the problem
            >>> problem = Problem(model="Maize", simulation="Simulation")
            >>> problem.add_control("Manager", "Sow using a variable rule", "Population", vtype="grid",
            ...                     start_value=5, values=[5, 9, 11])
            >>> problem.add_control("Manager", "Sow using a variable rule", "RowSpacing", vtype="grid",
            ...                     start_value=400, values=[400, 800, 1200])
            >>> result = problem.minimize_with_local_solver(method="Powell")
            >>> print(result.x_vars)
            {'Population': 11, 'RowSpacing': 800}
            """
        try:

            self._open_pbar(labels=[i.label for i in self.control_vars])
            call_counter = {"count": 0}
            func = getattr(self, '_set_objective_function')
            def wrapped_obj(x, *args):
                call_counter["count"] += 1
                self.pbar.update(1)
                return self._set_objective_function(x)

            wrapped = Objective(
                func= wrapped_obj,
                variables=self._variables
            )

            bounds = wrapped.bounds


            # Handle initial guess
            initial_guess = self.starting_values()

            encoded_initial = wrapped.encode(initial_guess)


            result = minimize(wrapped, x0=encoded_initial, bounds=bounds, **kwargs)

            # Attach labeled solution
            decoded_solution = wrapped.decode(result.x)
            result.x_vars = dict(zip(self.labels, decoded_solution))
            self.outcomes = result
            return result

        finally:
            self.clear_cache()
            self._close_pbar()

if __name__ == "__main__":
    maize_model = "Maize"
    class Problem(MixedVariableProblem):
        def __init__(self, model=None, simulation='Simulation'):
            super().__init__(model, simulation)
            self.simulation = simulation

        def evaluate(self, x, **kwargs):
            return -self.run(verbose=False).results.Yield.mean()


    problem = Problem(maize_model, simulation='Simulation')
    problem.add_control('Manager', "Sow using a variable rule", 'Population',vtype='grid',
                        start_value=5, values=[2,11, 12.5, 12, 10, 13,  5, 15])
    problem.add_control('Manager', "Sow using a variable rule", 'RowSpacing', vtype='grid', start_value=400,
                        values=[400, 600, 750, 800, 900, 1000, 1100, 1200])
    #de_res = problem.optimize_with_differential_evolution()
    res= problem.minimize_with_local_solver(method='Powell')




import asyncio
import argparse
import json
import os.path
import logging
import pprint
from apsimNGpy import settings
import time
import numpy as np
import pandas as pd
from apsimNGpy.core_utils.utils import timer
from apsimNGpy.core.apsim import ApsimModel, Models
from apsimNGpy.core.base_data import load_default_simulations
from apsimNGpy.settings import logger
from apsimNGpy.manager.weathermanager import get_weather, _is_within_USA_mainland
from apsimNGpy.cli.model_desc import extract_models
import os, sys
import asyncio
from apsimNGpy.cli.set_ups import print_msg, ColoredHelpFormatter
from apsimNGpy.manager.soilmanager import DownloadsurgoSoiltables, OrganiseSoilProfile


def print_msg(msg, normal = True):
    if normal:
       print(f'  \033[96m{msg}\033[0m')
    else:
        print(f'  \033[91m{msg}\033[0m')
def change_mgt(model, args):

    """
    Update management parameters in the model using a string specification of various parameters.

    The `args.management` string should contain one or more manager specifications in the form:
        'path=path_to_manager, param1=value, param2=value'

    To update multiple managers, separate them using a colon (`:`):
        'path=manager1, rate=20 : path=manager2, rate=30, start_day="2023-01-01"'

    Notes:
        - The 'path' key is required for each manager block.

    """

    if not args.management:
        return model

    def parse_management_args(management_str):
        parsed = []

        specs = management_str.split(':')
        for spec in specs:
            if ',' not in spec:
                raise argparse.ArgumentTypeError(f'invalide management specification {spec} detected')
            parts = [s.strip() for s in spec.split(',') if s.strip()]
            param_dict = {}
            for part in parts:
                if '__' in part:
                    raise argparse.ArgumentTypeError(f'invalid specification path `{part}` detected')
                if '=' not in part:
                    continue
                key, value = map(str.strip, part.split('=', 1))

                    # Evaluate non-path values (assumes they’re Python literals)
                param_dict[key] = value if key == 'path' else eval(value)
            if 'path' in param_dict:
                parsed.append(param_dict)
        return parsed

    for mgt_params in parse_management_args(args.management):
        print(mgt_params)
        model.update_mgt_by_path(**mgt_params)

    return model


async def fetch_weather_data(lonlat):
    """Fetch weather data asynchronously."""
    if _is_within_USA_mainland(lonlat):
        source = 'daymet'
    else:
        source = 'nasa'
    return await asyncio.to_thread(get_weather, lonlat=lonlat, source=source, start=1985, end=2020)


def replace_soil_from_web(lonlat):
    """
    Fetch soil data asynchronously
    @param lonlat:
    @return:
    """

    sp = DownloadsurgoSoiltables(lonlat)
    sop = OrganiseSoilProfile(sp, thickness=20, thickness_values=settings.SOIL_THICKNESS).cal_missingFromSurgo()
    return sop


async def run_apsim_model(model, report_name):
    print("running apsim")
    """Run APSIM model asynchronously."""
    # model.run(report_name)
    if report_name is None:
        report_name = model.inspect_model(Models.Report, fullpath=False)
    return await asyncio.to_thread(model.run, report_name=report_name, verbose=True)
    # return await asyncio.to_thread(model,)

async def create_experiment(model, factors, permutation =True, basename=None):
    print('creating experiment')
    if ";" in factors:
       factors = factors.split(';')
    else:
        factors = factors
    def _add(_model, _factors):
        for factor in _factors:
            _model.add_factor(factor) if factor else None
        return _model
    model.create_experiment(permutation=permutation, basename=basename)
    return _add(model, factors)

async def save_results(df, file_name):
    """Save results asynchronously."""
    await asyncio.to_thread(df.to_csv, file_name)


def clean(path):
    """Remove the apsim related files such as the .db files
    """


def replace_soil_data(model: ApsimModel, args):
    """Replace soil data"""
    s_nodes = ['chemical', 'physical', 'organic']
    for node in s_nodes:
        if getattr(args, node, None):
            valS = getattr(args, node)
            soil = valS.split(',')
            soil_eq = [soi.split('=') for soi in soil]
            _args_params = {k[0].strip(): k[1].strip() for k in
                            soil_eq}  # assumes that the second index 1 will be the value if we hand
            # "node_path='.Simulations.Simulation.Field.Soil.Physical, BD=1.23"

            for k, v in _args_params.items():
                if k != 'node_path':
                    _args_params[k] = eval(v)  # expect all other values not to be a string
                else:
                    nl = len(node)
                    if v[-nl:] != node.capitalize():
                        _args_params[k] = f"{v}.{node.capitalize()}"

            model.replace_soils_values_by_path(**_args_params)






soil_types_c2bare = {
    "Sand": 68,
    "Heavy clay": 73,
    "Loamy clay": 84,
    "Loam": 73,
    "Sandy loam": 68,
    "Sandy clay": 68,
    "Clay loam": 73,
    "Sandy clay loam": 73,
    "Loamy sand": 73,
    'Silty clay loam': 73,
     'Silt loam':73,
     'Silty clay': 68,
     'Fine sandy loam':68,
      'Loamy very fine sand':73,
      'Loamy fine sand': 73
}

for i in ['Clay loam', 'Silty clay loam', 'Sandy loam', 'Sandy clay loam',
       'Loam', 'Fine sandy loam', 'Loamy very fine sand',
       'Loamy fine sand', 'Loamy sand', 'Sand', 'Silt loam', 'Silty clay']:
    cl = soil_types_c2bare[i]
    print(cl)